from typing import Dict
from ..text import snake_case


BASE_URL = "https://raw.communitydragon.org/"


def merge_set_data(data: Dict, set: int, collection_key: str):
    set = int(max(data["sets"], key=int)) if set == -1 else set
    collected = {}
    for set_data in data["setData"]:
        if set_data["number"] != set:
            continue
        for item in set_data[collection_key]:
            collected[item["apiName"]] = item
    for item in data["sets"][str(set)][collection_key]:
        collected[item["apiName"]] = item
    return list(collected.values())


def abs_url(link: str, version="latest") -> str:
    '''Return the CDragon url for the given tft asset url'''
    if link is None: return link
    link = link.lower()
    if link[-3:] not in ["png","jpg","jpeg"]:
        link = link[:-3] + "png"
    return BASE_URL + version + "/game/" + link


def sanitize_item_description(string: str, obj: dict) -> str:
    '''Sanitize CDragon tft item descriptions'''
    if string is None: return string
    new_string = ""
    is_tag = False
    is_at = False
    is_percent = False
    percent = ""
    tag = ""
    at = ""
    for s in string:
        if not is_tag and not is_at and not is_percent and s not in "<>@": new_string += s
        if s == "<": is_tag = True
        elif s == ">":
            is_tag = False
            if tag == "br" and len(new_string) > 0 and new_string[-1] != " ": new_string += " "
            tag = ""
        elif s == "@" and not is_at: is_at = True
        elif s == "@":
            is_at = False
            try: new_string += str(obj[at])
            except KeyError: new_string += "(?)"
            at = ""
        elif s == "%": is_percent = True
        elif is_percent and s == " " and percent == "":
            is_percent = False
            new_string += s
        elif s == "%" and is_percent and len(percent) > 1:
            is_percent = False
            new_string = new_string[:-1]
        if is_tag and s not in "<>": tag += s
        if is_at and s != "@": at += s
        if is_percent and s != "%": percent += s
    return new_string


def sanitize_champion_description(string: str, list_of_obj: list) -> str:
    '''Sanitize CDragon tft champion descriptions'''
    if string is None: return string
    new_string = ""
    is_tag = False
    is_at = False
    tag = ""
    at = ""
    for s in string:
        if not is_tag and not is_at and s not in "<>@": new_string += s
        if s == "<": is_tag = True
        elif s == ">":
            is_tag = False
            if tag == "br" and len(new_string) > 0 and new_string[-1] != " ": new_string += " "
            tag = ""
        elif s == "@" and not is_at: is_at = True
        elif s == "@":
            is_at = False
            tags = snake_case(tag).split("_")
            found = False
            for t in tags:
                for obj in list_of_obj:
                    if t in snake_case(obj["name"]):
                        new_string += "/".join([str(vall) for vall in obj["value"][1:4]])
                        found = True
                        break
                if found: break
            if not found: new_string += "(?)"
            at = ""
        if is_tag and s not in "<>":
            tag += s
        if is_at and s != "@":
            at += s
    return new_string




from pyot.models import lor, riot
from pyot.utils.sync import async_to_sync

from .engine_core import assert_types, assert_walkable


@async_to_sync
async def test_card():
    o = await lor.Cards(set=5, version="latest", locale="en_us").get()
    assert_walkable(o)
    assert_types(o)


@async_to_sync
async def test_ranked():
    o = await lor.Leaderboard(region="americas").get()
    assert_walkable(o)
    assert_types(o)
    o = await lor.Leaderboard(region="sea").get()
    assert_walkable(o)
    assert_types(o)


@async_to_sync
async def test_match():
    a = await riot.Account(game_name="dango", tag_line="1557").using("lor").get()
    o = await lor.MatchHistory(puuid=a.puuid, region="sea").get()
    assert_walkable(o)
    assert_types(o)
    for ind in range(2):
        o_ = await o[ind].get()
        assert_walkable(o_)
        assert_types(o_)


@async_to_sync
async def test_status():
    o = await lor.Status(region="americas").get()
    assert_walkable(o)
    assert_types(o)



from abc import ABC, abstractproperty
from enum import Enum


class StoreType(Enum):
    SERVICE = "service"
    CACHE = "cache"


class Store(ABC):

    @property
    def classname(self):
        return self.__class__.__name__

    @abstractproperty
    def type(self) -> StoreType:
        raise NotImplementedError

    # Service store methods:

    async def request(self, method, token, *args, **kwargs):
        raise NotImplementedError

    async def get(self, token, *args, **kwargs):
        return await self.request("GET", token, *args, **kwargs)

    async def post(self, token, body, *args, **kwargs):
        return await self.request("POST", token, body, *args, **kwargs)

    async def put(self, token, body, *args, **kwargs):
        return await self.request("PUT", token, body, *args, **kwargs)

    # Cache store methods:

    async def set(self, token, response, *args, **kwargs):
        raise NotImplementedError

    async def clear(self, *args, **kwargs):
        raise NotImplementedError

    async def expire(self, *args, **kwargs):
        raise NotImplementedError

    async def delete(self, token, *args, **kwargs):
        raise NotImplementedError

    async def contains(self, token, *args, **kwargs):
        raise NotImplementedError



from time import time
from typing import Dict, List, Union

from aiohttp import ClientResponse

from pyot.utils.threading import AsyncLock

from .base import BaseLimiter, LimiterToken


class MemoryLimiterDict(dict):

    def get(self, k: str) -> int:
        try:
            return self.__getitem__(k)
        except KeyError:
            return None

    def set(self, k: str, v: int, nx=False):
        if nx and k not in self or not nx:
            self.__setitem__(k, v)
            return 1
        return 0

    def incr(self, k):
        self.__setitem__(k, super().get(k, 0) + 1)

    def decr(self, k):
        self.__setitem__(k, super().get(k, 0) - 1)


class MemoryLimiter(BaseLimiter):

    def __init__(self, game: str, api_key: str, limiting_share: int = 1):
        self._lock = AsyncLock()
        self.game = game
        self.api_key = api_key
        self.api_hash = api_key[-5:]
        self.limiting_share = limiting_share
        self.entries = MemoryLimiterDict()

    async def get_token(self, server: str, method: str):
        sleep = 0
        allowed = []
        pinging_list = []
        epoch = 0
        async with self._lock:
            app_prefix = f'{self.api_hash}:{self.game}:{server}'
            method_prefix = f'{self.api_hash}:{self.game}:{server}:{method}'
            now = epoch = time()
            for prefix, i, type in ((app_prefix, 0, "app"), (app_prefix, 1, "app"), (method_prefix, 0, "method"), (method_prefix, 1, "method")):
                prefix_i = f'{prefix}:{i}'
                exists = self.entries.get(f'{prefix_i}:exists')
                if exists == 0:
                    continue
                freeze = self.entries.get(f'{prefix_i}:freeze')
                if (freeze or 0) > now:
                    sleep = max(sleep, freeze - now)
                    continue
                pinging = self.entries.get(f'{prefix_i}:pinging')
                pingexp = self.entries.get(f'{prefix_i}:pingexp')
                if now > (pingexp or 0):
                    pinging = 0
                if pinging:
                    sleep = max(sleep, 0.1)
                    break
                maxcall = self.entries.get(f'{prefix_i}:maxcall')
                rollover = self.entries.get(f'{prefix_i}:rollover')
                timespan = self.entries.get(f'{prefix_i}:timespan')
                called = self.entries.get(f'{prefix_i}:called')
                begintime = self.entries.get(f'{prefix_i}:begintime')
                pingtime = self.entries.get(f'{prefix_i}:pingtime')
                if maxcall is not None:
                    blackout = begintime + timespan
                    nextstart = blackout + pingtime
                    qualified = nextstart - now
                if maxcall is None or now > nextstart:
                    self.entries.set(f'{prefix_i}:pinging', 1)
                    self.entries.set(f'{prefix_i}:pingexp', now + 10)
                    self.entries.set(f'{prefix_i}:pingbegintime', now)
                    pinging_list.append((prefix_i, type, i))
                    continue
                if (rollover or 0) + called >= maxcall:
                    sleep = max(sleep, qualified)
                    continue
                if now >= blackout:
                    sleep = max(sleep, qualified)
                    continue
                allowed.append(prefix_i)
            if sleep == 0:
                for allow in allowed:
                    self.entries.incr(f'{allow}:called')
                    if len(pinging_list) == 0:
                        self.entries.incr(f'{allow}:flying')
            else:
                for (prefix_i, type, _) in pinging_list:
                    self.entries.set(f'{prefix_i}:pinging', 0)
        return LimiterToken(server, method, epoch, sleep, allowed, pinging_list)

    async def sync_rates(self, token: LimiterToken, response: ClientResponse) -> Dict[str, List[List[int]]]:
        header = self.parse_headers(response)
        if header is None:
            await self.ping_fail(token)
            return
        if token.pinging:
            async with self._lock:
                now = time()
                for (prefix_i, type, i) in token.pinging:
                    if i >= len(header[f'{type}_limit']):
                        self.entries.set(f'{prefix_i}:exists', 0)
                    else:
                        self.entries.set(f'{prefix_i}:exists', 1)
                        self.entries.set(f'{prefix_i}:maxcall', header[f'{type}_limit'][i][0])
                        self.entries.set(f'{prefix_i}:timespan', header[f'{type}_limit'][i][1])
                        self.entries.set(f'{prefix_i}:called', header[f'{type}_count'][i][0])
                        self.entries.set(f'{prefix_i}:rollover', self.entries.get(f'{prefix_i}:flying'))
                        self.entries.set(f'{prefix_i}:flying', 0)
                        begintime = self.entries.get(f'{prefix_i}:pingbegintime')
                        self.entries.set(f'{prefix_i}:begintime', begintime)
                        self.entries.set(f'{prefix_i}:pingtime', now - begintime)
                        self.entries.set(f'{prefix_i}:pinging', 0)
        else:
            async with self._lock:
                for prefix_i in token.allowed:
                    begintime = self.entries.get(f'{prefix_i}:begintime')
                    if token.epoch >= begintime:
                        self.entries.decr(f'{prefix_i}:flying')
        return header

    async def ping_fail(self, token: LimiterToken):
        async with self._lock:
            for (prefix_i, _, _) in token.pinging:
                self.entries.set(f'{prefix_i}:pinging', 0)

    async def freeze_rates(self, token: LimiterToken, response: ClientResponse) -> Dict[str, Union[str, int]]:
        header = self.parse_429(response)
        now = time()
        async with self._lock:
            for (prefix_i, type, _) in token.pinging:
                if type != header["type"]:
                    continue
                self.entries.set(f'{prefix_i}:freeze', now + header["time"])
        return header



from typing import Mapping, Dict, Any
import asyncio

from pyot.pipeline.token import PipelineToken
from pyot.pipeline.handler import ErrorHandler
from pyot.endpoints.riotapi import RiotAPIEndpoint
from pyot.limiters.base import BaseLimiter
from pyot.utils.aiohttp import SafeClientSession
from pyot.utils.importlib import import_variable
from pyot.utils.logging import LazyLogger
from pyot.utils.safejson import loads
from pyot.utils.nullsafe import _

from .base import Store, StoreType


LOGGER = LazyLogger(__name__)


class RiotAPI(Store):

    type = StoreType.SERVICE

    def __init__(self, game: str, api_key: str, rate_limiter: Mapping[str, str] = None, error_handler: Mapping[int, Any] = None, log_level: int = 0, silence_429: bool = False):
        self.game = game
        self.api_key = api_key
        self.endpoints = RiotAPIEndpoint(game)
        self.handler = ErrorHandler(error_handler, 800)
        self.rate_limiter = self.create_rate_limiter(rate_limiter or {})
        self.log_level = log_level
        self.silence_429 = silence_429

    async def request(self, method: str, token: PipelineToken, body: Dict = None, session: SafeClientSession = None, **kwargs) -> Dict:
        url = self.endpoints.resolve(token)
        error_token = self.handler.get_token()
        while error_token.allow():
            limit_token = await self.rate_limiter.get_token(token.server, token.parent)
            if not limit_token.allow():
                await asyncio.sleep(limit_token.sleep)
                continue
            try:
                response = await session.request(method=method, url=url, headers={"X-Riot-Token": self.api_key}, json=body)
                LOGGER.log(self.log_level, f"[pyot.stores.riotapi:RiotAPI] {method} {token.value}")
            except Exception:
                response = None

            await self.rate_limiter.sync_rates(limit_token, response)
            status = _(response).status or 408
            if status == 200:
                return await response.json(encoding="utf-8", content_type=None, loads=loads)
            if status == 429:
                headers = await self.rate_limiter.freeze_rates(limit_token, response)
                if headers["type"] != "service" and not self.silence_429:
                    LOGGER.warning(f"[pyot.stores.riotapi:RiotAPI] A request failed with non-service 429, review confs if mass fails. Origin: {token.value}")
            await error_token.consume(status, token.value)

    def create_rate_limiter(self, dic: Dict[str, Any]) -> BaseLimiter:
        config = {key.lower(): val for (key, val) in dic.items()}
        try:
            limiter = import_variable(config.pop("backend"))
        except KeyError:
            limiter = import_variable('pyot.limiters.memory.MemoryLimiter')
        config["game"] = self.game
        config["api_key"] = self.api_key
        return limiter(**config)



from typing import List, TYPE_CHECKING

from pyot.conf.model import models
from pyot.core.functional import empty
from .base import PyotCore, PyotStatic

if TYPE_CHECKING:
    from .ranked import Leaderboard


# PYOT STATIC OBJECTS


class ContentLocalizedNamesData(PyotStatic):
    ar_ae: str
    de_de: str
    en_gb: str
    en_us: str
    es_es: str
    es_mx: str
    fr_fr: str
    id_id: str
    it_it: str
    ja_jp: str
    ko_kr: str
    pl_pl: str
    pt_br: str
    ru_ru: str
    th_th: str
    tr_tr: str
    vi_vn: str
    zh_cn: str
    zh_tw: str

    class Meta(PyotStatic.Meta):
        renamed = {
            "ar-ae": "ar_ae",
            "de-de": "de_de",
            "en-gb": "en_gb",
            "en-us": "en_us",
            "es-es": "es_es",
            "es-mx": "es_mx",
            "fr-fr": "fr_fr",
            "id-id": "id_id",
            "it-it": "it_it",
            "ja-jp": "ja_jp",
            "ko-kr": "ko_kr",
            "pl-pl": "pl_pl",
            "pt-br": "pt_br",
            "ru-ru": "ru_ru",
            "th-th": "th_th",
            "tr-tr": "tr_tr",
            "vi-vn": "vi_vn",
            "zh-cn": "zh_cn",
            "zh-tw": "zh_tw",
        }


class ContentItemData(PyotStatic):
    id: str
    name: str
    asset_name: str
    asset_path: str
    localized_names: ContentLocalizedNamesData


class ContentActData(PyotStatic):
    id: str
    name: str
    type: str
    parent_id: str
    localized_names: ContentLocalizedNamesData
    is_active: bool

    @property
    def leaderboard(self) -> "Leaderboard":
        from .ranked import Leaderboard
        return Leaderboard(act_id=self.id, platform=self.platform)


# PYOT CORE OBJECTS

class Content(PyotCore):
    version: str
    characters: List[ContentItemData]
    maps: List[ContentItemData]
    chromas: List[ContentItemData]
    skins: List[ContentItemData]
    skin_levels: List[ContentItemData]
    equips: List[ContentItemData]
    game_modes: List[ContentItemData]
    sprays: List[ContentItemData]
    spray_levels: List[ContentItemData]
    charms: List[ContentItemData]
    charm_levels: List[ContentItemData]
    player_cards: List[ContentItemData]
    player_titles: List[ContentItemData]
    ceremonies: List[ContentItemData]
    acts: List[ContentActData]

    class Meta(PyotCore.Meta):
        rules = {"content_v1_contents": []}

    def __init__(self, platform: str = models.val.DEFAULT_PLATFORM):
        self.initialize(locals())

    def query(self, locale: str = empty):
        '''Set query request parameters.'''
        super()._place_query(locals())
        return self



from abc import ABC, abstractmethod
from typing import Dict, List, Tuple

from aiohttp import ClientResponse

from pyot.utils.logging import LazyLogger


LOGGER = LazyLogger(__name__)


class LimiterToken:

    def __init__(self, server: str, method: str, epoch: float, sleep: float, allowed: List[str], pinging: List[Tuple[str, str, int]]) -> None:
        self.server = server
        self.method = method
        self.epoch = epoch
        self.sleep = sleep
        self.allowed = allowed
        self.pinging = pinging

    def allow(self):
        return self.sleep == 0


class BaseLimiter(ABC):

    def parse_headers(self, response: ClientResponse) -> Dict[str, List[List[int]]]:
        if response is None:
            return None
        headers = response.headers if hasattr(response, "headers") else None
        if headers is None:
            return None
        try:
            app_limit = [[int(val) for val in token.split(':')] for token in headers["X-App-Rate-Limit"].split(',')]
            app_count = [[int(val) for val in token.split(':')] for token in headers["X-App-Rate-Limit-Count"].split(',')]
            method_limit = [[int(val) for val in token.split(':')] for token in headers["X-Method-Rate-Limit"].split(',')]
            method_count = [[int(val) for val in token.split(':')] for token in headers["X-Method-Rate-Limit-Count"].split(',')]
            return {
                "app_limit": app_limit,
                "app_count": app_count,
                "method_limit": method_limit,
                "method_count": method_count,
            }
        except KeyError:
            return None

    def parse_429(self, response: ClientResponse) -> Dict:
        if response is None:
            return None
        headers = response.headers if hasattr(response, "headers") else None
        if headers is None:
            return None
        return {
            "type": response.headers["X-Rate-Limit-Type"] if "X-Rate-Limit-Type" in response.headers else "service",
            "time": float(response.headers["Retry-After"]) if "Retry-After" in response.headers else 1,
        }

    @abstractmethod
    def __init__(self, game: str, api_key: str, limiting_share: int = 1):
        self.game = game
        self.api_key = api_key
        self.limiting_share = limiting_share

    @abstractmethod
    async def get_token(self, server: str, method: str) -> LimiterToken:
        pass

    @abstractmethod
    async def sync_rates(self, token: LimiterToken, response: ClientResponse):
        pass

    @abstractmethod
    async def ping_fail(self, token: LimiterToken):
        pass

    @abstractmethod
    async def freeze_rates(self, token: LimiterToken, response: ClientResponse):
        pass



from datetime import datetime
from typing import List

from dateutil.parser import parse

from pyot.conf.model import models
from .base import PyotCore, PyotStatic

# PYOT STATIC OBJECTS

class StatusContentData(PyotStatic):
    locale: str
    content: str


class StatusUpdateData(PyotStatic):
    id: int
    author: str
    publish: bool
    publish_locations: List[str] # (Legal values: riotclient, riotstatus, game)
    created_at_strftime: str
    updated_at_strftime: str
    translations: List[StatusContentData]

    class Meta(PyotCore.Meta):
        raws = {"publish_locations"}
        renamed = {"created_at": "created_at_strftime", "updated_at": "updated_at_strftime"}

    @property
    def created_at(self) -> datetime:
        timestr = self.created_at_strftime
        return parse(timestr) if timestr is not None else timestr

    @property
    def updated_at(self) -> datetime:
        timestr = self.updated_at_strftime
        return parse(timestr) if timestr is not None else timestr


class StatusDetailData(PyotStatic):
    id: int
    maintenance_status: str # (Legal values: scheduled, in_progress, complete)
    incident_severity: str # (Legal values: info, warning, critical)
    created_at_strftime: str
    updated_at_strftime: str
    archive_at_strftime: str
    titles: List[StatusContentData]
    updates: List[StatusUpdateData]
    platforms: List[str]

    class Meta(PyotCore.Meta):
        raws = {"platforms"}
        renamed = {"created_at": "created_at_strftime", "updated_at": "updated_at_strftime", "archive_at": "archive_at_strftime"}

    @property
    def created_at(self) -> datetime:
        timestr = self.created_at_strftime
        return parse(timestr) if timestr is not None else timestr

    @property
    def archive_at(self) -> datetime:
        timestr = self.archive_at_strftime
        return parse(timestr) if timestr is not None else timestr

    @property
    def updated_at(self) -> datetime:
        timestr = self.updated_at_strftime
        return parse(timestr) if timestr is not None else timestr


# PYOT CORE OBJECTS

class Status(PyotCore):
    id: str
    name: str
    locales: List[str]
    maintenances: List[StatusDetailData]
    incidents: List[StatusDetailData]

    class Meta(PyotCore.Meta):
        rules = {"status_v1_platform_data": []}
        raws = {"locales"}

    def __init__(self, region: str = models.lor.DEFAULT_REGION):
        self.initialize(locals())



from typing import Any, TypeVar
from importlib import import_module


T = TypeVar("T")


def import_variable(path: str) -> Any:
    '''Return the class given its python path'''
    store_cls_name = path.split('.')[-1]
    store_path = '.'.join(path.split('.')[:-1])
    module = import_module(store_path)
    store_cls = getattr(module, store_cls_name)
    return store_cls



from datetime import datetime, timedelta

from pyot.models import lol
from pyot.utils.sync import async_to_sync
from pyot.utils.lol.routing import platform_to_region

from .engine_core import assert_types, assert_walkable


@async_to_sync
async def test_champion():
    o = await lol.Champion(id=81, locale="en_us").get()
    assert_walkable(o)
    assert_types(o)
    o = await lol.Champion(key="Irelia", locale="es_mx", version="12.1").get()
    assert_walkable(o)
    assert_types(o)
    o = await lol.Champions(version="pbe").get()
    assert_walkable(o)
    assert_types(o)


@async_to_sync
async def test_championmastery():
    s = await lol.Summoner(name="Morimorph", platform="na1").get()
    o = await lol.ChampionMasteries(summoner_id=s.id, platform="na1").get()
    assert_walkable(o)
    assert_types(o)
    o = await lol.ChampionMastery(summoner_id=s.id, champion_id=39, platform="na1").get()
    assert_walkable(o)
    assert_types(o)


@async_to_sync
async def test_championrotation():
    o = await lol.ChampionRotation(platform="na1").get()
    assert_walkable(o)
    assert_types(o)


@async_to_sync
async def test_clash():
    o = await lol.ClashTournaments(platform="na1").get()
    assert_walkable(o)
    assert_types(o)


@async_to_sync
async def test_item():
    o = await lol.Item(id=2003, locale="en_us").get()
    assert_walkable(o)
    assert_types(o)
    o = await lol.Items(locale="en_us").get()
    assert_walkable(o)
    assert_types(o)


@async_to_sync
async def test_league():
    o = await lol.ChallengerLeague(queue="RANKED_SOLO_5x5", platform="na1").get()
    assert_walkable(o)
    assert_types(o)
    o = await lol.GrandmasterLeague(queue="RANKED_SOLO_5x5", platform="na1").get()
    assert_walkable(o)
    assert_types(o)
    o = await lol.MasterLeague(queue="RANKED_SOLO_5x5", platform="na1").get()
    assert_walkable(o)
    assert_types(o)
    o = await lol.League(id=o.id, platform="na1").get()
    assert_walkable(o)
    assert_types(o)
    s = await lol.Summoner(name="Morimorph", platform="na1").get()
    o = await lol.SummonerLeague(summoner_id=s.id, platform="na1").get()
    assert_walkable(o)
    assert_types(o)
    o = await lol.DivisionLeague(tier="GOLD", division="I", queue="RANKED_SOLO_5x5", platform="na1").query(page=1).get()
    assert_walkable(o)
    assert_types(o)


@async_to_sync
async def test_match():
    s = await lol.Summoner(name="Morimorph", platform="na1").get()
    o = await lol.MatchHistory(puuid=s.puuid, region=platform_to_region(s.platform)).query(count=100, queue=420).get()
    assert_walkable(o)
    assert_types(o)
    o = await lol.MatchHistory(puuid=s.puuid, region=platform_to_region(s.platform)).query(count=100, queue=420, start_time=int((datetime.now() - timedelta(days=200)).timestamp())).get()
    assert_walkable(o)
    assert_types(o)
    o = await lol.MatchHistory(puuid=s.puuid, region=platform_to_region(s.platform)).query(count=100, queue=420, start_time=datetime.now() - timedelta(days=200)).get()
    assert_walkable(o)
    assert_types(o)
    o = await s.match_history.query(count=100, queue=420, start_time=datetime.now() - timedelta(days=200)).get()
    assert_walkable(o)
    assert_types(o)
    for ind in range(2):
        i = o[ind].id
        o_ = await o[ind].get()
        assert_walkable(o_)
        assert_types(o_)
        o1 = await lol.Match(id=i, region=platform_to_region(s.platform)).get()
        assert_walkable(o1)
        assert_types(o1)
        o2 = await lol.Timeline(id=i, region=platform_to_region(s.platform)).get()
        assert_walkable(o2)
        assert_types(o2)
        o1.feed_timeline(o2)
        assert_walkable(o1)
        assert_types(o1)


@async_to_sync
async def test_merakichampion():
    o = await lol.MerakiChampion(id=39).get()
    assert_walkable(o)
    assert_types(o)
    o = await lol.MerakiChampion(key="Irelia").get()
    assert_walkable(o)
    assert_types(o)
    o = await lol.MerakiChampion(name="Irelia").get()
    assert_walkable(o)
    assert_types(o)


@async_to_sync
async def test_merakiitem():
    o = await lol.MerakiItem(id=2003).get()
    assert_walkable(o)
    assert_types(o)


@async_to_sync
async def test_profileicon():
    o = await lol.ProfileIcons(locale="en_us").get()
    assert_walkable(o)
    assert_types(o)
    o = await o[230].get()
    assert_walkable(o)
    assert_types(o)


@async_to_sync
async def test_rune():
    o = await lol.Runes(locale="en_us").get()
    assert_walkable(o)
    assert_types(o)
    o = await lol.Rune(id=o[12].id, locale="en_us").get()
    assert_walkable(o)
    assert_types(o)


@async_to_sync
async def test_spectator():
    o = await lol.FeaturedGames(platform="euw1").get()
    assert_walkable(o)
    assert_types(o)
    s = await o[0].participants[0].summoner.get()
    o = await lol.CurrentGame(summoner_id=s.id, platform=s.platform).get()
    assert_walkable(o)
    assert_types(o)


@async_to_sync
async def test_spell():
    o = await lol.Spells(locale="en_us").get()
    assert_walkable(o)
    assert_types(o)
    o = await lol.Spell(id=12, locale="en_us", version="pbe").get()
    assert_walkable(o)
    assert_types(o)


@async_to_sync
async def test_status():
    o = await lol.Status(platform="na1").get()
    assert_walkable(o)
    assert_types(o)


@async_to_sync
async def test_summoner():
    o = await lol.Summoner(name="Morimorph").get()
    assert_walkable(o)
    assert_types(o)
    o = await lol.Summoner(puuid=o.puuid).get()
    assert_walkable(o)
    assert_types(o)



from typing import List, Iterator, TYPE_CHECKING

from pyot.conf.model import models
from pyot.core.functional import cache_indexes, lazy_property, empty
from pyot.utils.lol.cdragon import abs_url, sanitize_description
from .base import PyotCore

if TYPE_CHECKING:
    from .merakiitem import MerakiItem
    from .champion import Champion


# PYOT CORE OBJECT

class Item(PyotCore):
    id: int
    name: str
    description: str
    active: bool
    in_store: bool
    from_ids: List[int]
    to_ids: List[int]
    categories: List[str]
    maps: List[str]
    max_stacks: int
    modes: List[str]
    required_champion_key: str
    required_ally: str
    required_currency: str
    required_currency_cost: int
    is_enchantment: bool
    special_recipe_id: int
    self_cost: int
    total_cost: int
    icon_path: str

    class Meta(PyotCore.Meta):
        rules = {"cdragon_item_full": ["version", "locale", "?id"]}
        raws = {"from_ids", "to_ids", "categories", "maps", "modes"}
        renamed = {"from":"from_ids", "to": "to_ids", "map_string_id_inclusions": "maps", "mode_name_inclusions": "modes",
            "required_buff_currency_name": "required_currency", "required_buff_currency_cost": "required_currency_cost",
            "required_champion": "required_champion_key", "price": "self_cost", "price_total": "total_cost", "special_recipe": "special_recipe_id"}

    def __init__(self, id: int = empty, version: str = models.lol.DEFAULT_VERSION, locale: str = models.lol.DEFAULT_LOCALE):
        self.initialize(locals())

    @cache_indexes
    def filter(self, indexer, data):
        return indexer.get(self.id, data, "id")

    @lazy_property
    def icon_abspath(self) -> str:
        return abs_url(self.icon_path, self.metaroot.version)

    @lazy_property
    def cleaned_description(self) -> str:
        return sanitize_description(self.description)

    @property
    def from_items(self) -> List["Item"]:
        items = []
        for id in self.from_ids:
            items.append(Item(id=id, version=self.version, locale=self.locale))
        return items

    @property
    def to_items(self) -> List["Item"]:
        items = []
        for id in self.to_ids:
            items.append(Item(id=id, version=self.version, locale=self.locale))
        return items

    @property
    def meraki_item(self) -> "MerakiItem":
        from .merakiitem import MerakiItem
        return MerakiItem(id=self.id)

    @property
    def required_champion(self) -> "Champion":
        from .champion import Champion
        return Champion(key=self.required_champion_key, version=self.version, locale=self.locale)

    @property
    def special_recipe(self) -> "Item":
        return Item(id=self.special_recipe_id, version=self.version, locale=self.locale)


class Items(PyotCore):
    items: List[Item]

    class Meta(PyotCore.Meta):
        rules = {"cdragon_item_full": ["version", "locale"]}

    def __init__(self, version: str = models.lol.DEFAULT_VERSION, locale: str = models.lol.DEFAULT_LOCALE):
        self.initialize(locals())

    def __getitem__(self, item):
        if not isinstance(item, int):
            return super().__getitem__(item)
        return self.items[item]

    def __iter__(self) -> Iterator[Item]:
        return iter(self.items)

    def __len__(self):
        return len(self.items)

    def transform(self, data):
        return {"items": data}



from .league import ChallengerLeague, MasterLeague, GrandmasterLeague, DivisionLeague, SummonerLeague, League
from .profileicon import ProfileIcon, ProfileIcons
from .thirdpartycode import ThirdPartyCode
from .champion import Champion, Champions
from .match import MatchHistory, Match
from .trait import Trait, Traits
from .summoner import Summoner
from .item import Item, Items


from typing import List, Dict, Iterator, TYPE_CHECKING

from pyot.conf.model import models
from pyot.core.functional import lazy_property, empty
from pyot.utils.lol.champion import id_by_key, id_by_name
from pyot.utils.lol.cdragon import abs_url, sanitize_description
from .base import PyotCore, PyotStatic

if TYPE_CHECKING:
    from .merakichampion import MerakiChampion


# PYOT STATIC OBJECTS

class ChampionTacticalData(PyotStatic):
    style: int
    difficulty: int
    damage_type: str


class ChampionPlayerStyleData(PyotStatic):
    damage: int
    durability: int
    crowd_control: int
    mobility: int
    utility: int


class ChampionChromaDescriptionsData(PyotStatic):
    region: str = None
    description: str


class ChampionChromaRaritiesData(PyotStatic):
    region: str = None
    description: str
    rarity: int


class ChampionSkinChromaData(PyotStatic):
    id: int
    name: str
    chroma_path: str
    colors: List[str]
    descriptions: List[ChampionChromaDescriptionsData]
    rarities: List[ChampionChromaRaritiesData]

    class Meta(PyotStatic.Meta):
        raws = {"colors"}

    @lazy_property
    def chroma_abspath(self) -> str:
        return abs_url(self.chroma_path, self.metaroot.version)


class ChampionSkinData(PyotStatic):
    id: int
    is_base: bool
    name: str
    splash_path: str
    uncentered_splash_path: str
    tile_path: str
    load_screen_path: str
    load_screen_vintage_path: str
    skin_type: str
    rarity: str
    is_legacy: bool
    chroma_path: str
    chromas: List[ChampionSkinChromaData]
    emblems: List[str]
    skin_line: int
    description: str
    splash_video_path: str
    collection_splash_video_path: str
    features_text: str
    region_rarity_id: int
    rarity_gem_path: str

    class Meta(PyotStatic.Meta):
        raws = {"emblems"}
        renamed = {"skin_lines": "skin_line"}

    @lazy_property
    def splash_abspath(self) -> str:
        return abs_url(self.splash_path, self.metaroot.version)

    @lazy_property
    def uncentered_splash_abspath(self) -> str:
        return abs_url(self.uncentered_splash_path, self.metaroot.version)

    @lazy_property
    def tile_abspath(self) -> str:
        return abs_url(self.tile_path, self.metaroot.version)

    @lazy_property
    def load_screen_abspath(self) -> str:
        return abs_url(self.load_screen_path, self.metaroot.version)

    @lazy_property
    def load_screen_vintage_abspath(self) -> str:
        return abs_url(self.load_screen_vintage_path, self.metaroot.version)

    @lazy_property
    def chroma_abspath(self) -> str:
        return abs_url(self.chroma_path, self.metaroot.version)

    @lazy_property
    def splash_video_abspath(self) -> str:
        return abs_url(self.splash_video_path, self.metaroot.version)

    @lazy_property
    def collection_splash_video_abspath(self) -> str:
        return abs_url(self.collection_splash_video_path, self.metaroot.version)


class ChampionPassiveData(PyotStatic):
    name: str
    icon_path: str
    description: str
    ability_video_path: str
    ability_video_image_path: str

    class Meta(PyotStatic.Meta):
        renamed = {"ability_icon_path": "icon_path"}

    @lazy_property
    def icon_abspath(self) -> str:
        return abs_url(self.icon_path, self.metaroot.version)


class ChampionSpellData(PyotStatic):
    key: str
    name: str
    icon_path: str
    cost: List[float]
    cooldown: List[float]
    range: List[float]
    description: str
    long_description: str
    ability_video_path: str
    ability_video_image_path: str
    max_level: int
    formulas: Dict
    coefficients: Dict[str, float]
    effect_amounts: Dict[str, List[float]]
    ammo: Dict[str, List[float]]

    class Meta(PyotStatic.Meta):
        raws = {"cost", "cooldown", "range", "formulas", "coefficients", "effect_amounts", "ammo"}
        renamed = {"spell_key": "key", "ability_icon_path": "icon_path", "dynamic_description": "long_description"}

    @lazy_property
    def icon_abspath(self) -> str:
        return abs_url(self.icon_path, self.metaroot.version)

    @lazy_property
    def cleaned_description(self) -> str:
        return sanitize_description(self.long_description)


class ChampionAbilityData(PyotStatic):
    p: ChampionPassiveData
    q: ChampionSpellData
    w: ChampionSpellData
    e: ChampionSpellData
    r: ChampionSpellData


# PYOT CORE OBJECTS

class Champion(PyotCore):
    id: int
    key: str
    name: str
    lore: str
    tactical_info: ChampionTacticalData
    play_style: ChampionPlayerStyleData
    square_path: str
    stinger_sfx_path: str
    choose_vo_path: str
    ban_vo_path: str
    roles: List[str]
    skins: List[ChampionSkinData]
    abilities: ChampionAbilityData
    passive: ChampionPassiveData
    title: str
    recommended_item_defaults: List[str]

    class Meta(PyotCore.Meta):
        rules = {
            "cdragon_champion_by_id": ["version", "locale", "id"],
        }
        raws = {"roles", "recommended_item_defaults"}
        renamed = {"alias": "key", "short_bio": "lore", "playstyle_info": "play_style", "square_portrait_path": "square_path", "spells": "abilities"}

    def __init__(self, id: int = empty, key: str = empty, name: str = empty, version: str = models.lol.DEFAULT_VERSION, locale: str = models.lol.DEFAULT_LOCALE):
        self.initialize(locals())

    async def setup(self):
        if not hasattr(self, "id"):
            if hasattr(self, "key"):
                self.id = await id_by_key(self.key)
            elif hasattr(self, "name"):
                self.id = await id_by_name(self.name)

    def transform(self, data):
        if not data.get("skins", None):
            return data
        for skin in data["skins"]:
            if skin["skinLines"] is not None:
                skin["skinLines"] = skin["skinLines"][0]["id"]
        spells = {}
        for spell in data["spells"]:
            spell["cost"] = spell.pop("costCoefficients")
            spell["cooldown"] = spell.pop("cooldownCoefficients")
            spells[spell["spellKey"]] = spell
        spells["p"] = data["passive"].copy()
        data["spells"] = spells
        return data

    @lazy_property
    def square_abspath(self) -> str:
        return abs_url(self.square_path, self.metaroot.version)

    @lazy_property
    def stinger_sfx_abspath(self) -> str:
        return abs_url(self.stinger_sfx_path, self.metaroot.version)

    @lazy_property
    def choose_vo_abspath(self) -> str:
        return abs_url(self.choose_vo_path, self.metaroot.version)

    @lazy_property
    def ban_vo_abspath(self) -> str:
        return abs_url(self.ban_vo_path, self.metaroot.version)

    @property
    def meraki_champion(self) -> "MerakiChampion":
        from .merakichampion import MerakiChampion
        return MerakiChampion(id=self.id)


class Champions(PyotCore):
    champions: List[Champion]

    class Meta(PyotCore.Meta):
        rules = {"cdragon_champion_summary": ["version", "locale"]}

    def __init__(self, version: str = models.lol.DEFAULT_VERSION, locale: str = models.lol.DEFAULT_LOCALE):
        self.initialize(locals())

    def __getitem__(self, item):
        if not isinstance(item, int):
            return super().__getitem__(item)
        return self.champions[item]

    def __iter__(self) -> Iterator[Champion]:
        return iter(self.champions)

    def __len__(self):
        return len(self.champions)

    def transform(self, data):
        return {"champions": data}



from datetime import datetime
from typing import TYPE_CHECKING

from pyot.conf.model import models
from pyot.core.functional import empty
from .base import PyotCore

if TYPE_CHECKING:
    from .league import SummonerLeague
    from .thirdpartycode import ThirdPartyCode
    from .profileicon import ProfileIcon
    from ..riot.account import Account
    from .match import MatchHistory


# PYOT CORE OBJECTS

class Summoner(PyotCore):
    name: str
    id: str
    account_id: str
    level: int
    puuid: str
    profile_icon_id: int
    revision_date_millis: int

    class Meta(PyotCore.Meta):
        rules = {
            "summoner_v1_by_id": ["id"],
            "summoner_v1_by_account_id": ["account_id"],
            "summoner_v1_by_puuid": ["puuid"],
            "summoner_v1_by_name": ["name"],
        }
        renamed = {"summoner_level": "level", "revision_date": "revision_date_millis"}

    def __init__(self, id: str = empty, account_id: str = empty, name: str = empty, puuid: str = empty, platform: str = models.tft.DEFAULT_PLATFORM):
        self.initialize(locals())

    @property
    def revision_date(self) -> datetime:
        return datetime.fromtimestamp(self.revision_date_millis//1000)

    @property
    def league_entries(self) -> "SummonerLeague":
        from .league import SummonerLeague
        return SummonerLeague(summoner_id=self.id, platform=self.platform)

    @property
    def third_party_code(self) -> "ThirdPartyCode":
        from .thirdpartycode import ThirdPartyCode
        return ThirdPartyCode(summoner_id=self.id, platform=self.platform)

    @property
    def profile_icon(self) -> "ProfileIcon":
        from .profileicon import ProfileIcon
        return ProfileIcon(id=self.profile_icon_id)

    @property
    def account(self) -> "Account":
        from ..riot.account import Account
        return Account(puuid=self.puuid).using(self.metapipeline.name)

    @property
    def match_history(self) -> "MatchHistory":
        from .match import MatchHistory
        return MatchHistory(puuid=self.puuid, region=self.region)



from typing import List, Iterator, TYPE_CHECKING, NoReturn

from pyot.conf.model import models
from pyot.core.functional import empty
from .base import PyotCore, PyotStatic

if TYPE_CHECKING:
    from .summoner import Summoner


# PYOT STATIC OBJECTS

class MiniSeriesData(PyotStatic):
    target: int
    wins: int
    losses: int
    progress: str


class LeagueEntryData(PyotStatic):
    summoner_id: str
    summoner_name: str
    league_points: int
    rank: str
    wins: int
    losses: int
    veteran: bool
    inactive: bool
    fresh_blood: bool
    hot_streak: bool
    mini_series: MiniSeriesData

    @property
    def summoner(self) -> "Summoner":
        from .summoner import Summoner
        return Summoner(id=self.summoner_id, name=self.summoner_name, platform=self.platform)


class SummonerLeagueEntryData(LeagueEntryData):
    league_id: str
    queue: str
    tier: str

    class Meta(LeagueEntryData.Meta):
        renamed = {"queue_type": "queue"}

    @property
    def league(self) -> "League":
        return League(id=self.league_id, platform=self.platform)


# PYOT CORE OBJECTS

class League(PyotCore):
    tier: str
    id: str
    queue: str
    name: str
    entries: List[LeagueEntryData]

    class Meta(PyotCore.Meta):
        rules = {"league_v4_league_by_league_id": ["id"]}
        renamed = {"league_id": "id"}

    def __init__(self, id: str = empty, platform: str = models.lol.DEFAULT_PLATFORM):
        self.initialize(locals())


class ApexLeague(League):
    queue: str
    id: str

    class Meta(League.Meta):
        pass

    def __init__(self, queue: str = empty, platform: str = models.lol.DEFAULT_PLATFORM):
        self.initialize(locals())

    @property
    def league(self) -> League:
        return League(id=self.id, platform=self.platform)


class ChallengerLeague(ApexLeague):

    class Meta(ApexLeague.Meta):
        rules = {"league_v4_challenger_league": ["queue"]}


class GrandmasterLeague(ApexLeague):

    class Meta(ApexLeague.Meta):
        rules = {"league_v4_grandmaster_league": ["queue"]}


class MasterLeague(ApexLeague):

    class Meta(ApexLeague.Meta):
        rules = {"league_v4_master_league": ["queue"]}


class SummonerLeague(PyotCore):
    summoner_id: str
    entries: List[SummonerLeagueEntryData]

    class Meta(PyotCore.Meta):
        rules = {"league_v4_summoner_entries": ["summoner_id"]}

    def __init__(self, summoner_id: str = empty, platform: str = models.lol.DEFAULT_PLATFORM):
        self.initialize(locals())

    def __getitem__(self, item):
        if not isinstance(item, int):
            return super().__getitem__(item)
        return self.entries[item]

    def __iter__(self) -> Iterator[SummonerLeagueEntryData]:
        return iter(self.entries)

    def __len__(self):
        return len(self.entries)

    def transform(self, data):
        new_data = {}
        new_data["entries"] = data
        return new_data

    @property
    def summoner(self) -> "Summoner":
        from .summoner import Summoner
        return Summoner(id=self.summoner_id, platform=self.platform)


class DivisionLeague(SummonerLeague):
    queue: str
    division: str
    tier: str

    class Meta(SummonerLeague.Meta):
        rules = {"league_v4_entries_by_division": ["queue", "tier", "division"]}

    def __init__(self, queue: str = empty, division: str = empty, tier: str = empty, platform: str = models.lol.DEFAULT_PLATFORM):
        self.initialize(locals())

    def query(self, page: int = empty):
        '''Set query request parameters.'''
        super()._place_query(locals())
        return self

    @property
    def summoner(self) -> NoReturn:
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute 'summoner'")



from typing import List, TYPE_CHECKING
from datetime import datetime

from pyot.conf.model import models
from pyot.core.functional import empty
from .base import PyotCore, PyotStatic

if TYPE_CHECKING:
    from .summoner import Summoner


# PYOT STATIC OBJECTS

class ClashPlayerData(PyotStatic):
    summoner_id: str
    team_id: str
    position: str
    role: str

    @property
    def summoner(self) -> "Summoner":
        from .summoner import Summoner
        return Summoner(id=self.summoner_id, platform=self.platform)

    @property
    def team(self) -> "ClashTeam":
        return ClashTeam(id=self.team_id, platform=self.platform)


class ClashTournamentPhaseData(PyotStatic):
    id: int
    registration_timestamp: int
    start_timestamp: int
    cancelled: bool

    class Meta(PyotStatic.Meta):
        renamed = {'start_time': 'start_timestamp', 'registration_time': 'registration_timestamp'}

    @property
    def start_time(self) -> datetime:
        return datetime.fromtimestamp(self.start_timestamp//1000)

    @property
    def registration_time(self) -> datetime:
        return datetime.fromtimestamp(self.registration_timestamp//1000)


class ClashTournamentData(PyotStatic):
    id: int
    theme_id: int
    name_key: str
    name_key_secondary: str
    schedule: List[ClashTournamentPhaseData]


# PYOT CORE OBJECTS

class ClashPlayers(PyotCore):
    summoner_id: str
    players: List[ClashPlayerData]

    class Meta(PyotCore.Meta):
        rules = {"clash_v1_players": ["summoner_id"]}

    def __init__(self, summoner_id: str = empty, platform: str = models.lol.DEFAULT_PLATFORM):
        self.initialize(locals())

    def transform(self, data):
        new_data = {}
        new_data["players"] = data
        return new_data

    @property
    def summoner(self) -> "Summoner":
        from .summoner import Summoner
        return Summoner(id=self.summoner_id, platform=self.platform)


class ClashTeam(PyotCore):
    id: str
    tournament_id: int
    name: str
    icon_id: int
    tier: int
    captain_summoner_id: str
    abbreviation: str
    players: List[ClashPlayerData]

    class Meta(PyotCore.Meta):
        renamed = {"captain": "captain_summoner_id"}
        rules = {"clash_v1_teams": ["id"]}

    def __init__(self, id: str = empty, platform: str = models.lol.DEFAULT_PLATFORM):
        self.initialize(locals())

    def transform(self, data):
        for player in data["players"]:
            player["teamId"] = self.id
        return data

    @property
    def captain(self) -> "Summoner":
        from .summoner import Summoner
        return Summoner(id=self.captain_summoner_id, platform=self.platform)

    @property
    def tournament(self) -> "ClashTournament":
        return ClashTournament(id=self.tournament_id, platform=self.platform)


class ClashTournaments(PyotCore):
    tournaments: List[ClashTournamentData]

    class Meta(PyotCore.Meta):
        rules = {"clash_v1_tournaments_all": []}

    def __init__(self, platform: str = models.lol.DEFAULT_PLATFORM):
        self.initialize(locals())

    def __getitem__(self, name):
        if not isinstance(name, int):
            return super().__getitem__(name)
        return self.tournaments[name]

    def __iter__(self) -> List[ClashTournamentData]:
        return iter(self.tournaments)

    def __len__(self):
        return len(self.tournaments)

    def transform(self, data):
        new_data = {}
        new_data["tournaments"] = data
        return new_data


class ClashTournament(ClashTournamentData, PyotCore):
    team_id: str

    class Meta(PyotCore.Meta):
        rules = {
            "clash_v1_tournaments_by_team_id": ["team_id"],
            "clash_v1_toutnaments_by_tournament_id": ["id"]
        }

    def __init__(self, id: int = empty, team_id: str = empty, platform: str = models.lol.DEFAULT_PLATFORM):
        self.initialize(locals())

    @property
    def team(self) -> ClashTeam:
        return ClashTeam(id=self.team_id, platform=self.platform)
    



from .base import BaseEndpoint


class DDragonEndpoint(BaseEndpoint):

    base_url = "https://dd.b.pvp.net"

    all = {
        "lor": {
            "ddragon_lor_set_data": "/{version}/set{set}/{locale}/data/set{set}-{locale}.json"
        }
    }



import asyncio
from typing import Dict
import copy

from pyot.core import exceptions as exc


class ErrorHandler:

    DEFAULT_HANDLER = {
        204 : ("T", []),
        400 : ("T", []),
        401 : ("T", []),
        404 : ("T", []),
        403 : ("T", []),
        405 : ("T", []),
        408 : ("E", [3, 3]),
        429 : ("E", [3, 3]),
        500 : ("E", [3, 3]),
        502 : ("E", [3, 3]),
        503 : ("E", [3, 3]),
        504 : ("E", [3, 3]),
        602 : ("R", [2]),
        800 : ("T", []),
    }

    def __init__(self, handler: Dict, on_key_error: int = 800) -> None:
        self.handler = copy.deepcopy(self.DEFAULT_HANDLER)
        self.on_key_error = on_key_error
        if handler is None:
            return
        self.validate_handler(handler)
        self.handler.update(handler)

    def __getitem__(self, code: int):
        try:
            return self.handler[code]
        except KeyError:
            return self.handler[self.on_key_error]

    def validate_handler(self, handler: Dict):
        for code, c in handler.items():
            if code not in self.DEFAULT_HANDLER:
                raise RuntimeError(f"Handler code {code} is not a valid status for error handling")
            strategy = c[0]
            try:
                params = c[1]
            except IndexError as e:
                raise AttributeError("Handler values receives 2 paramameters, 1 was given") from e
            for a in params:
                if not isinstance(a, int):
                    raise RuntimeError(f"Handler strategy takes 'int' parameters, '{type(a)}' was given")
            if strategy == "T":
                if len(params) != 0:
                    raise RuntimeError(f"Handler strategy 'T' (Throw) takes 0 parameters, but {len(params)} was given")
            elif strategy == "E":
                if len(params) != 2:
                    raise RuntimeError(f"Handler strategy 'E' (Exp. backoff) takes 2 parameters, but {len(params)} was given")
            elif strategy == "R":
                if len(params) != 1:
                    raise RuntimeError(f"Handler strategy 'R' (Retry) takes 1 parameters, but {len(params)} was given")
            else:
                raise RuntimeError(f"Handler strategy '{strategy}' is not a valid token for error handling")

    def get_token(self) -> "ErrorToken":
        return ErrorToken(self)


class ErrorToken:

    def __init__(self, instance: ErrorHandler, tries: int = 0, raise_at: int = 1, exception: Exception = None, code: int = 0) -> None:
        self.instance = instance
        self.tries = tries
        self.raise_at = raise_at
        self.exception = exception
        self.code = code

    async def consume(self, code: int, origin: str):
        how = self.instance[code]
        strategy = how[0]
        params = how[1]

        if self.code != code:
            self.exception = None
            self.code = code

        if self.exception is None:
            if code == 404:
                self.exception = exc.NotFound(origin)
            elif code == 204:
                self.exception = exc.NoContent(origin)
            elif code in [500, 502, 503, 504]:
                self.exception = exc.ServerError(code, origin)
            elif code == 429:
                self.exception = exc.RateLimited(origin)
            elif code == 403:
                self.exception = exc.Forbidden(origin)
            elif code == 405:
                self.exception = exc.MethodNotAllowed(origin)
            elif code == 401:
                self.exception = exc.Unauthorized(origin)
            elif code == 400:
                self.exception = exc.BadRequest(origin)
            elif code == 408:
                self.exception = exc.Timeout(origin)
            elif code == 602:
                self.exception = exc.DecodeError(origin)
            elif code == 601:
                self.exception = exc.SessionNotFound(origin)
            elif code == 600:
                self.exception = exc.NotFindable()
            elif code == 800:
                self.exception = exc.UnidentifiedResponse(code, origin)
            else:
                self.exception = Exception("Unexpected error. Please open an issue on github repository")

            if strategy != "T":
                self.raise_at = params[-1]+1
            else:
                self.raise_at = 1

        self.tries += 1
        if strategy == "E" and self.tries < self.raise_at:
            await asyncio.sleep(params[0]**self.tries)

    def allow(self):
        if self.tries < self.raise_at:
            return True
        raise self.exception



from abc import ABC, abstractproperty
from typing import Dict

from pyot.core import exceptions as exc
from pyot.pipeline.token import PipelineToken


class BaseEndpoint(ABC):

    @abstractproperty
    def base_url(self) -> str:
        raise NotImplementedError

    shared: Dict[str, str] = {}
    all: Dict[str, Dict[str, str]] = {}

    def __init__(self, game: str):
        try:
            self.endpoints = self.all[game]
        except KeyError as e:
            name = self.__class__.__name__.replace("Endpoint", "")
            raise NotImplementedError(f"{name} does not support '{e}' model") from e
        self.endpoints.update(self.shared)

    def resolve(self, token: PipelineToken) -> str:
        try:
            url = (self.base_url + self.endpoints[token.method]).format(**self.clean(token))
            query = self.query(token.queries)
            return url + query
        except KeyError as e:
            raise exc.NotFindable from e

    def query(self, queries: Dict) -> str:
        query = ""
        for a, b in queries.items():
            if isinstance(b, list):
                for val in b:
                    query = query + "&" + str(a) + "=" + str(val)
            else:
                query = query + "&" + str(a) + "=" + str(b)
        if len(query) > 1:
            query = "?" + query[1:]
        return query

    def clean(self, token: PipelineToken) -> Dict[str, str]:
        return token.params



import datetime
from typing import Any
import pytz

import motor.motor_asyncio as mongo
from pyot.core.exceptions import NotFound
from pyot.pipeline.token import PipelineToken
from pyot.pipeline.expiration import ExpirationManager
from pyot.core.resources import ResourceTemplate
from pyot.utils.logging import LazyLogger

from .base import Store, StoreType


LOGGER = LazyLogger(__name__)


class MongoDB(Store):

    type = StoreType.CACHE

    def __init__(self, game: str, db: str, host='127.0.0.1', port=27017, expirations: Any = None, log_level: int = 0, **kwargs) -> None:
        self.game = game
        if 'connect' not in kwargs:
            kwargs['connect'] = False
        if 'w' not in kwargs:
            kwargs['w'] = 0
        self.db = db
        self.initialized = False
        self.mongos = ResourceTemplate(
            acquire_func=lambda: self.acquire_database(host=host, port=port, **kwargs),
            release_func=lambda database: database.client.close()
        )
        self.location = f"{host}:{port}:{db}"
        self.expirations = ExpirationManager(game, expirations)
        self.log_level = log_level

    async def acquire_database(self, host: str, port: int, **kwargs):
        database = mongo.AsyncIOMotorClient(host=host, port=port, **kwargs)[self.db]
        for method in self.expirations:
            indexes = await database[method].index_information()
            if 'setAt_1' in indexes:
                if self.expirations.get_timeout(method) >= 0:
                    await database.command('collMod', method, index={'keyPattern': {'setAt': 1}, 'expireAfterSeconds': self.expirations.get_timeout(method)})
                else:
                    await database[method].drop_index('setAt')
            elif self.expirations.get_timeout(method) > 0:
                await database[method].create_index('token')
                await database[method].create_index('setAt', expireAfterSeconds=self.expirations.get_timeout(method))
            else:
                await database[method].create_index('token')
        return database

    async def set(self, token: PipelineToken, value: Any, **kwargs) -> None:
        timeout = self.expirations.get_timeout(token.method)
        if timeout != 0:
            cache = await self.mongos.acquire()
            await cache[token.method].insert_one(
                {
                    'token': token.value,
                    'data': value,
                    'dataType': "bson",
                    'setAt': datetime.datetime.now(pytz.utc)
                }
            )
            LOGGER.log(self.log_level, f"[pyot.stores.mongodb:MongoDB#{self.location}] SET {token.value}")

    async def get(self, token: PipelineToken, **kwargs) -> Any:
        timeout = self.expirations.get_timeout(token.method)
        if timeout == 0:
            raise NotFound(token.value)
        cache = await self.mongos.acquire()
        item = await cache[token.method].find_one({'token': token.value})
        if item is None:
            raise NotFound(token.value)
        LOGGER.log(self.log_level, f"[pyot.stores.mongodb:MongoDB#{self.location}] GET {token.value}")
        if item.get("dataType", None) is None:
            await cache[token.method].delete_many({'token': token.value})
            raise NotFound(token.value)
        return item["data"]

    async def delete(self, token: PipelineToken, **kwargs) -> None:
        cache = await self.mongos.acquire()
        await cache[token.method].delete_many({'token': token.value})
        LOGGER.log(self.log_level, f"[pyot.stores.mongodb:MongoDB#{self.location}] DELETE {token.value}")

    async def contains(self, token: PipelineToken, **kwargs) -> bool:
        cache = await self.mongos.acquire()
        item = await cache[token.method].find_one({'token': token.value})
        if item is None:
            return False
        return True

    async def clear(self, **kwargs):
        cache = await self.mongos.acquire()
        collections = await cache.list_collection_names()
        for name in collections:
            await cache.drop_collection(name)
        LOGGER.log(self.log_level, f"[pyot.stores.mongodb:MongoDB#{self.location}] Store cleared successfully")



import inspect
from typing import Any, List, Type


def get_method_properties(clas: Type[object]) -> List[str]:
    return inspect.getmembers(clas, predicate=lambda func: (
        (
            inspect.isfunction(func) and
            func.__name__ in clas.__dict__
        ) or (
            isinstance(func, property) and
            func.fget and
            func.fget.__name__ in clas.__dict__
        ) or (
            inspect.ismethod(func) and
            func.__name__ in clas.__dict__ and
            isinstance(clas.__dict__[func.__name__], (classmethod, staticmethod))
        )
    ))


def get_method_property_names(clas: Any) -> List[str]:
    return [i[0] for i in inspect.getmembers(clas, predicate=lambda func: inspect.isfunction(func) or isinstance(func, property))]


def newline_join(any):
    return '\n'.join(any)



import asyncio
import functools
import warnings

from aiohttp import ClientSession


warnings.filterwarnings('ignore', category=DeprecationWarning)


class SafeClientSession(ClientSession):

    def _create_closed_event(self) -> asyncio.Event:
        """
        Work around aiohttp issue that doesn't properly close transports on exit.
        See https://github.com/aio-libs/aiohttp/issues/1925#issuecomment-639080209
        """

        transports = 0
        all_is_lost = asyncio.Event()

        def connection_lost(exc, orig_lost):
            nonlocal transports
            try:
                orig_lost(exc)
            finally:
                transports -= 1
                if transports == 0:
                    all_is_lost.set()

        def eof_received(orig_eof_received):
            try:
                orig_eof_received()
            except AttributeError:
                # eof_received() is called after _app_protocol and _transport are set to None
                pass

        for conn in self.connector._conns.values():
            for handler, _ in conn:
                proto: asyncio.Protocol = getattr(handler.transport, "_ssl_protocol", None)
                if proto is None:
                    continue
                transports += 1
                orig_lost = proto.connection_lost
                orig_eof_received = proto.eof_received
                proto.connection_lost = functools.partial(
                    connection_lost, orig_lost=orig_lost
                )
                proto.eof_received = functools.partial(
                    eof_received, orig_eof_received=orig_eof_received
                )

        if transports == 0:
            all_is_lost.set()

        return all_is_lost

    async def close(self) -> None:
        '''Graceful close and release all resources.'''
        closed_event = self._create_closed_event()
        await super().close()
        await closed_event.wait()


warnings.filterwarnings('default', category=DeprecationWarning)



from typing import List, Iterator, TYPE_CHECKING

from pyot.conf.model import models
from pyot.core.functional import lazy_property, cache_indexes, empty
from pyot.core.exceptions import NotFound
from pyot.utils.tft.cdragon import merge_set_data, sanitize_champion_description, abs_url
from .base import PyotCore, PyotStatic

if TYPE_CHECKING:
    from .trait import Trait


# PYOT STATIC OBJECTS

class ChampionAbilityVariableData(PyotStatic):
    name: str
    value: List[float]

    class Meta(PyotStatic.Meta):
        raws = {"value"}


class ChampionAbilityData(PyotStatic):
    name: str
    description: str
    icon_path: str
    variables: List[ChampionAbilityVariableData]

    class Meta(PyotStatic.Meta):
        renamed = {"desc": "description", "icon": "icon_path"}

    @lazy_property
    def icon_abspath(self) -> str:
        return abs_url(self.icon_path, self.metaroot.version)

    @lazy_property
    def cleaned_description(self) -> str:
        return sanitize_champion_description(self.description, self["variables"])


class ChampionStatData(PyotStatic):
    armor: float
    attack_speed: float
    crit_chance: float
    crit_multiplier: float
    damage: float
    hp: float
    initial_mana: float
    magic_resist: float
    mana: float
    range: float


# PYOT CORE OBJECTS

class Champion(PyotCore):
    set: int
    key: str
    name: str
    cost: int
    stats: ChampionStatData
    trait_keys: List[str]
    ability: ChampionAbilityData
    lol_id: int
    icon_path: str

    class Meta(PyotCore.Meta):
        raws = {"trait_keys"}
        rules = {"cdragon_tft_full": ["?key", "?set", "version", "locale"]}
        renamed = {"api_name": "key", "traits": "trait_keys", "icon": "icon_path"}

    def __init__(self, key: str = empty, set: int = empty, version: str = models.tft.DEFAULT_VERSION, locale: str = models.tft.DEFAULT_LOCALE):
        self.initialize(locals())
        if key and set is None:
            self.find_set()

    def find_set(self):
        try:
            self.set = int(self.key.split("_")[0][3:])
        except Exception:
            self.set = -1

    @cache_indexes
    def filter(self, indexer, data):
        return indexer.get(
            self.key,
            merge_set_data(data, self.set, "champions"),
            "apiName"
        )

    @lazy_property
    def icon_abspath(self) -> str:
        return abs_url(self.icon_path, self.metaroot.version)

    @property
    def traits(self) -> List["Trait"]:
        from .trait import Trait
        return [Trait(key=i, set=self.set, version=self.version, locale=self.locale) for i in self.trait_keys]


class Champions(PyotCore):
    set: int
    champions: List[Champion]

    class Meta(PyotCore.Meta):
        rules = {"cdragon_tft_full": ["?set", "version", "locale"]}

    def __init__(self, set: int = -1, version: str = models.tft.DEFAULT_VERSION, locale: str = models.tft.DEFAULT_LOCALE):
        self.initialize(locals())

    def __getitem__(self, item):
        if not isinstance(item, int):
            return super().__getitem__(item)
        return self.champions[item]

    def __iter__(self) -> Iterator[Champion]:
        return iter(self.champions)

    def __len__(self):
        return len(self.champions)

    def filter(self, data):
        try:
            return merge_set_data(data, self.set, "champions")
        except KeyError as e:
            raise NotFound("Request was successful but filtering gave no matching item") from e

    def transform(self, data):
        return {"champions": data}



from typing import Tuple, Dict

from pyot.pipeline.token import PipelineToken
from pyot.pipeline.handler import ErrorHandler
from pyot.endpoints.ddragon import DDragonEndpoint
from pyot.utils.aiohttp import SafeClientSession
from pyot.utils.logging import LazyLogger
from pyot.utils.safejson import loads
from pyot.utils.nullsafe import _

from .base import Store, StoreType


LOGGER = LazyLogger(__name__)


class DDragon(Store):

    type = StoreType.SERVICE

    def __init__(self, game: str, error_handler: Dict[int, Tuple] = None, log_level: int = 0):
        self.game = game
        self.handler = ErrorHandler(error_handler, 800)
        self.endpoints = DDragonEndpoint(game)
        self.log_level = log_level

    async def get(self, token: PipelineToken, session: SafeClientSession, **kwargs) -> Dict:
        url = self.endpoints.resolve(token)
        error_token = self.handler.get_token()
        while error_token.allow():
            try:
                response = await session.request("GET", url)
                LOGGER.log(self.log_level, f"[pyot.stores.ddragon:DDragon] GET {token.value}")
            except Exception:
                response = None

            status = _(response).status or 408
            if status == 200:
                return await response.json(encoding="utf-8", content_type=None, loads=loads)
            await error_token.consume(status, token.value)



from typing import List, TYPE_CHECKING

from pyot.conf.model import models

from .base import PyotCore

if TYPE_CHECKING:
    from .champion import Champion
    from .merakichampion import MerakiChampion


# PYOT CORE OBJECTS

class ChampionRotation(PyotCore):
    free_champion_ids: List[int]
    free_newie_champion_ids: List[int]
    newie_max_level: int

    class Meta(PyotCore.Meta):
        rules = {"champion_v3_rotation": []}
        raws = {"free_champion_ids", "free_newie_champion_ids"}
        renamed = {"free_champion_ids_for_new_players": "free_newie_champion_ids", "max_new_player_level": "newie_max_level"}

    def __init__(self, platform: str = models.lol.DEFAULT_PLATFORM):
        self.initialize(locals())

    @property
    def free_champions(self) -> List["Champion"]:
        from .champion import Champion
        mutable = []
        for i in self.free_champion_ids:
            mutable.append(Champion(id=i))
        return mutable

    @property
    def free_newie_champions(self) -> List["Champion"]:
        from .champion import Champion
        mutable = []
        for i in self.free_newie_champion_ids:
            mutable.append(Champion(id=i))
        return mutable

    @property
    def meraki_free_champions(self) -> List["MerakiChampion"]:
        from .merakichampion import MerakiChampion
        mutable = []
        for i in self.free_champion_ids:
            mutable.append(MerakiChampion(id=i))
        return mutable

    @property
    def meraki_free_newie_champions(self) -> List["MerakiChampion"]:
        from .merakichampion import MerakiChampion
        mutable = []
        for i in self.free_newie_champion_ids:
            mutable.append(MerakiChampion(id=i))
        return mutable



import asyncio
import inspect
import sys
import traceback
from types import TracebackType
from typing import Any, Callable, List, Dict, Awaitable, TypeVar, Optional, Type

from pyot.utils.logging import LazyLogger


LOGGER = LazyLogger(__name__)

T = TypeVar('T')

ExceptionHandlerFunction = Callable[[Type[BaseException], BaseException, TracebackType], Any]


class QueueItem:

    def __init__(self, id, coro):
        self.id = id
        self.coro = coro


class Queue:
    '''A managed Queue on top of asyncio.Queue. This Queue is only usable as a context manager.'''
    _id_counter: int
    _queue: asyncio.Queue

    workers: int
    maxsize: int
    responses: Dict
    worker_tasks: List[asyncio.Task]
    exception_handler: Callable[[Exception], Any]

    def __init__(self, workers: int = 25, maxsize: int = None, log_level: int = 0, exception_handler: ExceptionHandlerFunction = traceback.print_exception):
        if workers < 1: raise ValueError('Number of workers must be an integer greater than 0')
        self.workers = workers
        self.exception_handler = exception_handler
        if maxsize is None:
            self.maxsize = workers * 2
        else:
            self.maxsize = maxsize
        self.log_level = log_level

    async def worker(self, queue: asyncio.Queue):
        while True:
            item: QueueItem = await queue.get()
            try:
                res = await item.coro
                if res is not None:
                    self.responses[item.id] = res
            except Exception:
                one, two, three = sys.exc_info()
                self.exception_handler(one, two, three)
            finally:
                queue.task_done()

    async def __aenter__(self) -> "Queue":
        self._queue = asyncio.Queue(maxsize=self.maxsize)
        self.responses = {}
        self.worker_tasks = []
        self._id_counter = 0
        for _ in range(self.workers):
            worker = asyncio.create_task(self.worker(self._queue))
            self.worker_tasks.append(worker)
        LOGGER.log(self.log_level, f"[pyot.core.queue:Queue] Spawned {self.workers} workers")
        return self

    async def __aexit__(self, *args):
        await self._queue.join()
        for worker in self.worker_tasks:
            worker.cancel()
        await asyncio.gather(*self.worker_tasks, return_exceptions=True)
        LOGGER.log(self.log_level, f"[pyot.core.queue:Queue] Joined {self.workers} workers")

    async def put(self, coro: Awaitable, delay: float = 0):
        '''
        Put a coroutine object to the queue, if the queue is full, wait for availability.
        A delay may be provided if desired for execution balancing.
        '''
        if delay > 0:
            await asyncio.sleep(delay)
        if not inspect.isawaitable(coro):
            raise ValueError(f"{str(coro)} is not awaitable")
        await self._queue.put(QueueItem(self._id_counter, coro))
        self._id_counter += 1

    async def join(self, class_of_t: Optional[Type[T]] = None) -> List[T]:
        '''
        Block until all items in the queue have been gotten and processed.
        Empty the collected responses and returns them.

        `None` and `Exception` responses are not collected, order of the responses might not correspond the put order.
        '''
        await self._queue.join()
        response = [res[1] for res in sorted(self.responses.items())]
        self.responses = {}
        self._id_counter = 0
        return response





BASE_URL = "https://raw.communitydragon.org/"


def strip_k(string: str) -> str:
    '''Strips char k if string start with k'''
    if string is None: return string
    return string[1:] if string[0] == "k" else string


def abs_url(link: str, version="latest") -> str:
    '''Return the CDragon url for the given game asset url'''
    if link is None: return link
    link = link.lower()
    splited = link.split("/lol-game-data/assets/")
    if len(splited) == 2:
        return BASE_URL + version + "/plugins/rcp-be-lol-game-data/global/default/" + splited[1]
    return BASE_URL + version + "/plugins/rcp-be-lol-game-data/global/default/" + link


def sanitize_description(string: str) -> str:
    '''Sanitize CDragon descriptions'''
    if string is None: return string
    new_string = ""
    is_tag = False
    is_at = False
    tag = ""
    for s in string:
        if not is_tag and not is_at and s not in "<>@": new_string += s
        if s == "<": is_tag = True
        elif s == ">":
            is_tag = False
            if tag == "br" and len(new_string) > 0 and new_string[-1] != " ": new_string += "\n"
            tag = ""
        elif s == "@" and not is_at:
            is_at = True
            new_string += "(?)"
        elif s == "@": is_at = False
        if is_tag and s not in "<>": tag += s
    return new_string



from typing import List, TYPE_CHECKING, Union

from pyot.utils.lol.champion import key_by_id, key_by_name
from pyot.core.functional import empty
from .base import PyotStatic, PyotCore

if TYPE_CHECKING:
    from .champion import Champion


# PYOT STATIC OBJECTS

class MerakiChampionSpellModifierData(PyotStatic):
    values: List[float]
    units: List[str]

    class Meta(PyotStatic.Meta):
        raws = {"values", "units"}


class MerakiChampionSpellAttrData(PyotStatic):
    attribute: str
    modifiers: List[MerakiChampionSpellModifierData]
    affected_by_cdr: bool


class MerakiChampionSpellEffectData(PyotStatic):
    description: str
    leveling: List[MerakiChampionSpellAttrData]


class MerakiChampionSpellData(PyotStatic):
    name: str
    icon: str
    effects: List[MerakiChampionSpellEffectData]
    cost: MerakiChampionSpellAttrData
    cooldown: MerakiChampionSpellAttrData
    targeting: str
    affects: str
    spellshieldable: str
    resource: str
    damage_type: str
    spell_effects: str
    projectile: str
    on_hit_effects: str
    occurrence: int
    notes: str
    blurb: str
    missile_speed: str
    recharge_rate: str
    collision_radius: str
    tether_radius: str
    on_target_cd_static: str
    inner_radius: str
    speed: str
    width: str
    angle: str
    cast_time: str
    effect_radius: str
    target_range: str


class MerakiChampionAbilityData(PyotStatic):
    p: List[MerakiChampionSpellData]
    q: List[MerakiChampionSpellData]
    w: List[MerakiChampionSpellData]
    e: List[MerakiChampionSpellData]
    r: List[MerakiChampionSpellData]


class MerakiChampionAttributeRatingData(PyotStatic):
    damage: int
    toughness: int
    control: int
    mobility: int
    utility: int
    ability_reliance: int
    attack: int
    defense: int
    magic: int
    difficulty: int


class MerakiChampionStatDetailData(PyotStatic):
    flat: float
    percent: float
    per_level: float
    percent_per_level: float


class MerakiChampionStatData(PyotStatic):
    health: MerakiChampionStatDetailData
    health_regen: MerakiChampionStatDetailData
    mana: MerakiChampionStatDetailData
    mana_regen: MerakiChampionStatDetailData
    armor: MerakiChampionStatDetailData
    magic_resistance: MerakiChampionStatDetailData
    attack_damage: MerakiChampionStatDetailData
    movespeed: MerakiChampionStatDetailData
    acquisition_radius: MerakiChampionStatDetailData
    selection_radius: MerakiChampionStatDetailData
    pathing_radius: MerakiChampionStatDetailData
    gameplay_radius: MerakiChampionStatDetailData
    critical_strike_damage: MerakiChampionStatDetailData
    critical_strike_damage_modifier: MerakiChampionStatDetailData
    attack_speed: MerakiChampionStatDetailData
    attack_speed_ratio: MerakiChampionStatDetailData
    attack_cast_time: MerakiChampionStatDetailData
    attack_total_time: MerakiChampionStatDetailData
    attack_delay_offset: MerakiChampionStatDetailData
    attack_range: MerakiChampionStatDetailData
    aram_damage_taken: MerakiChampionStatDetailData
    aram_damage_dealt: MerakiChampionStatDetailData
    aram_healing: MerakiChampionStatDetailData
    aram_shielding: MerakiChampionStatDetailData
    urf_damage_taken: MerakiChampionStatDetailData
    urf_damage_dealt: MerakiChampionStatDetailData
    urf_healing: MerakiChampionStatDetailData
    urf_shielding: MerakiChampionStatDetailData


class MerakiChampionPriceData(PyotStatic):
    blue_essence: int
    rp: int
    sale_rp: int


class MerakiChampionChromaDescriptionsData(PyotStatic):
    region: str = None
    description: str


class MerakiChampionChromaRaritiesData(PyotStatic):
    region: str = None
    description: str
    rarity: int


class MerakiChampionSkinChromaData(PyotStatic):
    id: int
    name: str
    chroma_path: str
    colors: List[str]
    descriptions: List[MerakiChampionChromaDescriptionsData]
    rarities: List[MerakiChampionChromaRaritiesData]

    class Meta(PyotStatic.Meta):
        raws = {"colors"}


class MerakiChampionSkinData(PyotStatic):
    name: str
    id: int
    is_base: bool
    availability: str
    format_name: str
    loot_eligible: bool
    cost: Union[str, int]
    sale: int
    distribution: str
    rarity: str
    chromas: List[MerakiChampionSkinChromaData]
    lore: str
    release: str
    set: List[str]
    splash_path: str
    uncentered_splash_path: str
    tile_path: str
    load_screen_path: str
    load_screen_vintage_path: str
    new_effects: bool
    new_animations: bool
    new_recall: bool
    new_voice: bool
    new_quotes: bool
    voice_actor: List[str]
    splash_artist: List[str]

    class Meta(PyotStatic.Meta):
        raws = {"set", "voice_actor", "splash_artist"}


# PYOT CORE OBJECTS

class MerakiChampion(PyotCore):
    id: int
    key: str
    name: str
    title: str
    full_name: str
    icon: str
    resource: str
    attack_type: str
    adaptive_type: str
    stats: MerakiChampionStatData
    roles: List[str]
    attribute_ratings: MerakiChampionAttributeRatingData
    abilities: MerakiChampionAbilityData
    release_date: str
    release_patch: str
    patch_last_changed: str
    price: MerakiChampionPriceData
    skins: List[MerakiChampionSkinData]
    lore: str

    class Meta(PyotCore.Meta):
        server_type = "locale"
        rules = {"meraki_champion_by_key": ["key"]}
        raws = {"roles"}

    def __init__(self, id: int = empty, key: str = empty, name: str = empty):
        self.initialize({"locale": "default", **locals()})

    async def setup(self):
        if not hasattr(self, "key"):
            if hasattr(self, "id"):
                self.key = await key_by_id(self.id)
            elif hasattr(self, "name"):
                self.key = await key_by_name(self.name)
        if self.key == "FiddleSticks":  # MERAKI HAS LOWERCASE S ON FIDDLE
            self.key = "Fiddlesticks"

    @property
    def champion(self) -> "Champion":
        from .champion import Champion
        return Champion(key=self.key, locale="en_us")

    def transform(self, data):
        if data["key"] == "Fiddlesticks":  # MERAKI HAS LOWERCASE S ON FIDDLE, GO BACK
            data["key"] = "FiddleSticks"
        return data



from typing import Any, Generic, TypeVar, Union


T = TypeVar("T")


class NullSafe:
    '''Documentation at: <https://github.com/paaksing/nullsafe-python>'''

    def __getattr__(self, k: str) -> "NullSafe":
        return undefined

    def __getitem__(self, k: str) -> "NullSafe":
        return undefined

    def __bool__(self):
        return False

    def __eq__(self, o: object) -> bool:
        if o is None or o is undefined or isinstance(o, NullSafe):
            return True
        return False

    def __call__(self, *args: Any, **kwds: Any) -> "NullSafe":
        return undefined

    def __repr__(self) -> str:
        return "undefined"

    def __str__(self) -> str:
        return "undefined"

    def __setattr__(self, name: str, value: Any) -> None:
        raise AttributeError(f"'{self.__class__.__name__}' object can't set attribute")

    def __iter__(self):
        raise TypeError(f"'{repr(self)}' object is not iterable")


undefined = NullSafe()


class NullSafeProxy(Generic[T]):

    __o: T

    def __init__(self, o: T) -> None:
        self.__o = o

    def __getitem__(self, k: str) -> Union[Any,  NullSafe]:
        try:
            val = self.__o.__getitem__(k)
            if val is None:
                return undefined
            return val
        except (KeyError, AttributeError):
            return undefined

    def __getattr__(self, name: str) -> Union[Any,  NullSafe]:
        val = getattr(self.__o, name, undefined)
        return undefined if val is None else val

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "_NullSafeProxy__o":
            return super().__setattr__(name, value)
        raise AttributeError(f"'{self.__class__.__name__}' object can't set attribute")

    def __repr__(self) -> str:
        return f"{self.__class__.__module__}.{self.__class__.__name__}({repr(self.__o)})"

    def __str__(self) -> str:
        return str(self.__o)


def nullsafe(o: T) -> Union[T, NullSafe, NullSafeProxy[T]]:
    if o == None:
        return undefined
    return NullSafeProxy(o)


_ = nullsafe



from threading import Lock
import asyncio


class AsyncLock:
    '''
    An asynchronous threading Lock. The event loop won't be blocked when acquiring the lock.
    '''
    def __init__(self):
        self._lock = Lock()

    async def __aenter__(self, *args) -> bool:
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(None, self._lock.acquire)

    async def __aexit__(self, *args):
        self._lock.release()

    async def acquire(self) -> bool:
        '''Acquire the lock without locking the loop'''
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(None, self._lock.acquire)

    def release(self):
        '''Release the lock, this is not async because it is immediate and useful for hooks (e.g. registering `atexit`)'''
        self._lock.release()



from typing import TypeVar
import pickle


T = TypeVar("T")


def fast_copy(obj: T) -> T:
    '''30x faster copy than `copy.deepcopy`, but not all objects can be fast copied (e.g. lambdas).'''
    return pickle.loads(pickle.dumps(obj, protocol=-1))



from pyot.core.objects import PyotCoreBase, PyotStaticBase
from pyot.conf.pipeline import pipelines


MODULE_REPR = 'League of Legends'


class PyotRouting:

    _regions = {"americas", "europe", "asia", "sea", "esports"}
    _platforms = {"br1", "eun1", "euw1", "jp1", "kr", "la1", "la2", "na1", "oc1", "tr1", "ru", "ph2", "sg2", "th2", "tw2", "vn2"}
    _platform2regions = {
        "br1": "americas",
        "eun1": "europe",
        "euw1": "europe",
        "jp1": "asia",
        "kr": "asia",
        "la1": "americas",
        "la2": "americas",
        "na1": "americas",
        "oc1": "sea",
        "tr1": "europe",
        "ru": "europe",
        "ph2": "sea",
        "sg2": "sea",
        "th2": "sea",
        "tw2": "sea",
        "vn2": "sea",
    }


class PyotCore(PyotRouting, PyotCoreBase):

    class Meta(PyotCoreBase.Meta):
        pipeline = pipelines.lol


class PyotStatic(PyotRouting, PyotStaticBase):

    class Meta(PyotStaticBase.Meta):
        pipeline = pipelines.lol


bases = (PyotCore, PyotStatic)



from pyot.core.objects import PyotCoreBase, PyotStaticBase
from pyot.conf.pipeline import pipelines


MODULE_REPR = 'Valorant'


class PyotRouting:

    _regions = {"americas", "europe", "asia", "esports"}
    _platforms = {"ap", "br", "esports", "eu", "kr", "latam", "na"}


class PyotCore(PyotRouting, PyotCoreBase):

    class Meta(PyotCoreBase.Meta):
        pipeline = pipelines.val


class PyotStatic(PyotRouting, PyotStaticBase):

    class Meta(PyotStaticBase.Meta):
        pipeline = pipelines.val


bases = (PyotCore, PyotStatic)



#!/usr/bin/env python

import sys
from os import path

from setuptools import setup, find_packages

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


install_requires = [
    "wheel",
    "python-dateutil>=2.8",
    "aiohttp>=3.8",
    "pytz",
    "lor-deckcodes",
]

extras_require = {
    "diskcache": ["diskcache>=5.1", "asgiref>=3.2"],
    "redis": ["aioredis[hiredis]>=2.0"],
    "mongodb": ["motor>=2.3"],
    "test": ["typeguard>=2.13", "aioredis[hiredis]>=2.0", "motor>=2.3", "diskcache>=5.1", "asgiref>=3.2"],
}


# Require python 3.7
if sys.version_info.major != 3 and sys.version_info.minor < 7:
    sys.exit("'Pyot' requires Python >= 3.7")

setup(
    name="pyot",
    version="6.0.6",
    author="Paaksing",
    author_email="paaksingtech@gmail.com",
    url="https://github.com/paaksing/Pyot",
    description="AsyncIO based high level Python framework Riot Games API framework which encourages rapid development and clean, pragmatic design.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=["Riot Games", "League of Legends", "Teamfight Tactics", "Valorant", "Legends of Runeterra", "API", "REST", "asyncio"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3.7",
        "Environment :: Web Environment",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Topic :: Games/Entertainment",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Natural Language :: English",
    ],
    license="MIT",
    packages=find_packages(exclude=("test")),
    zip_safe=True,
    install_requires=install_requires,
    extras_require=extras_require,
    include_package_data=True,
)



from typing import List, Iterator, TYPE_CHECKING

from pyot.core.functional import empty
from .base import PyotCore, PyotStatic

if TYPE_CHECKING:
    from .summoner import Summoner


# PYOT STATIC

class TournamentLobbyEventData(PyotStatic):
    summoner_id: str
    event_type: str
    timestamp: str


# PYOT CORE

class TournamentProvider(PyotCore):
    id: int
    region: str

    class Meta(PyotCore.Meta):
        rules = {"tournament_v4_providers": []}

    def __init__(self, region: str = empty):
        self.initialize(locals())

    def body(self, region: str, url: str):
        '''Set body request parameters.'''
        super()._place_body(locals())
        return self

    def validate(self):
        if not hasattr(self._meta, "body"):
            raise TypeError("This object's body parameters is required")

    def transform(self, data):
        return {"id": data}


class TournamentStubProvider(TournamentProvider):

    class Meta(TournamentProvider.Meta):
        rules = {"tournament_stub_v4_providers": []}


class Tournament(PyotCore):
    id: int
    region: str

    class Meta(PyotCore.Meta):
        rules = {"tournament_v4_tournaments": []}

    def __init__(self, region: str = empty):
        self.initialize(locals())

    def body(self, name: str, provider_id: int):
        '''Set body request parameters.'''
        super()._place_body(locals())
        return self

    def validate(self):
        if not hasattr(self._meta, "body"):
            raise TypeError("This object's body parameters is required")

    def transform(self, data):
        return {"id": data}


class TournamentStub(Tournament):

    class Meta(Tournament.Meta):
        rules = {"tournament_stub_v4_tournaments": []}


class TournamentLobbyEvents(PyotCore):
    events: List[TournamentLobbyEventData]
    region: str

    class Meta(PyotCore.Meta):
        rules = {"tournament_v4_lobby_events": ["code"]}

    def __init__(self, code: int = empty, region: str = empty):
        self.initialize(locals())

    def transform(self, data):
        return {"events": data}

    def __getitem__(self, item):
        if not isinstance(item, int):
            return super().__getitem__(item)
        return self.events[item]

    def __iter__(self) -> Iterator[TournamentLobbyEventData]:
        return iter(self.events)

    def __len__(self):
        return len(self.events)


class TournamentStubLobbyEvents(TournamentLobbyEvents):

    class Meta(TournamentLobbyEvents.Meta):
        rules = {"tournament_stub_v4_lobby_events": ["code"]}


class TournamentCode(PyotCore):
    code: str
    spectators: str
    lobby_name: str
    meta_data: str
    password: str
    team_size: int
    provider_id: int
    pick_type: str
    tournament_id: int
    id: int
    map: str
    hosted_region: str
    summoner_ids: List[str]

    class Meta(PyotCore.Meta):
        raws = {"summoner_ids"}
        renamed = {"region": "hosted_region", "participants": "summoner_ids"}
        rules = {"tournament_v4_codes_by_code": ["code"]}

    def __init__(self, code: str = empty, region: str = empty):
        self.initialize(locals())

    def body(self, map_type: str, pick_type: str, spectator_type: str, allowed_summoner_ids: List[str] = empty):
        '''Set body request parameters.'''
        super()._place_body(locals())
        return self

    @property
    def summoners(self) -> "Summoner":
        from .summoner import Summoner
        return [Summoner(id=id_) for id_ in self.summoner_ids]


class TournamentCodes(PyotCore):
    codes: List[str]
    region: str

    class Meta(PyotCore.Meta):
        raws = {"codes"}
        rules = {"tournament_v4_codes": []}

    def __init__(self, region: str = empty):
        self.initialize(locals())

    def query(self, tournament_id: int, count: int = empty):
        '''Set query request parameters.'''
        super()._place_query(locals())
        return self

    def body(self, map_type: str, pick_type: str, team_size: int, spectator_type: str, allowed_summoner_ids: List[str] = None, metadata: str = None):
        '''Set body request parameters.'''
        super()._place_body(locals())
        return self

    def validate(self):
        if not hasattr(self._meta, "body"):
            raise TypeError("This object's body parameters is required")
        if not hasattr(self._meta, "query"):
            raise TypeError("This object is missing query parameters")

    def transform(self, data):
        return {"codes": data}

    @property
    def tournament_codes(self) -> List[TournamentCode]:
        return [TournamentCode(code=code, region=self.region) for code in self.codes]


class TournamentStubCodes(TournamentCodes):

    class Meta(TournamentCodes.Meta):
        rules = {"tournament_stub_v4_codes": []}



from typing import Dict
from datetime import timedelta as td
import copy
import warnings

from pyot.core.warnings import PyotPipelineWarning


class ExpirationManager:

    shared: Dict[str, int] = {
        "account_v1_by_puuid": 0,
        "account_v1_by_riot_id": 0,
        "account_v1_active_shard": 0,
    }

    all: Dict[str, Dict] = {
        "lol": {
            "champion_v3_rotation": 0,
            "champion_mastery_v4_all_mastery": 0,
            "champion_mastery_v4_by_champion_id": 0,
            "clash_v1_players": 0,
            "clash_v1_teams": 0,
            "clash_v1_tournaments_by_team_id": 0,
            "clash_v1_toutnaments_by_tournament_id": 0,
            "clash_v1_tournaments_all": 0,
            "league_v4_summoner_entries": 0,
            "league_v4_challenger_league": 0,
            "league_v4_grandmaster_league": 0,
            "league_v4_master_league": 0,
            "league_v4_entries_by_division": 0,
            "league_v4_league_by_league_id": 0,
            "status_v4_platform_data": 0,
            "match_v4_match": 0,
            "match_v4_timeline": 0,
            "match_v4_matchlist": 0,
            "match_v4_tournament_match": 0,
            "match_v4_tournament_matches": 0,
            "match_v5_match": 0,
            "match_v5_timeline": 0,
            "match_v5_matches": 0,
            "spectator_v4_current_game": 0,
            "spectator_v4_featured_games": 0,
            "summoner_v4_by_name": 0,
            "summoner_v4_by_id": 0,
            "summoner_v4_by_account_id": 0,
            "summoner_v4_by_puuid": 0,
            "third_party_code_v4_code": 0,
            "tournament_v4_codes_by_code": 0,
            "tournament_v4_lobby_events": 0,
            "tournament_stub_v4_lobby_events": 0,

            "cdragon_champion_by_id": td(minutes=20),
            "cdragon_champion_summary": td(minutes=20),
            "cdragon_item_full": td(minutes=20),
            "cdragon_rune_full": td(minutes=20),
            "cdragon_profile_icon_full": td(minutes=20),
            "cdragon_spells_full": td(minutes=20),

            "meraki_champion_by_key": td(minutes=20),
            "meraki_item_by_id": td(minutes=20),
        },
        "lor": {
            "ranked_v1_leaderboards": 0,
            "match_v1_matchlist": 0,
            "match_v1_match": 0,
            "ddragon_lor_set_data": td(minutes=20),
            "status_v1_platform_data": 0,
        },
        "tft": {
            "league_v1_summoner_entries": 0,
            "league_v1_challenger_league": 0,
            "league_v1_grandmaster_league": 0,
            "league_v1_master_league": 0,
            "league_v1_entries_by_division": 0,
            "league_v1_league_by_league_id": 0,
            "match_v1_match": 0,
            "match_v1_matchlist": 0,
            "summoner_v1_by_name": 0,
            "summoner_v1_by_id": 0,
            "summoner_v1_by_account_id": 0,
            "summoner_v1_by_puuid": 0,

            "cdragon_tft_full": td(minutes=20),
            "cdragon_profile_icon_full": td(minutes=20),
        },
        "val": {
            "match_v1_match": 0,
            "match_v1_matchlist": 0,
            "match_v1_recent": 0,
            "content_v1_contents": 0,
            "ranked_v1_leaderboards": 0,
            "status_v1_platform_data": 0,
        }
    }

    def __init__(self, game: str, custom_expirations: Dict):
        self.expirations = copy.deepcopy(self.all[game])
        self.expirations.update(copy.deepcopy(self.shared))
        if custom_expirations is not None:
            for key in custom_expirations:
                if key not in self.expirations:
                    raise RuntimeError(f"'{key}' is not a valid expiration token")
            self.expirations.update(custom_expirations)
        self.expirations = self.create_expiration(self.expirations)

    def create_expiration(self, expirations):
        expirations_ = {}
        for key, time in expirations.items():
            if isinstance(time, td):
                expirations_[key] = int(time.total_seconds())
            else:
                try:
                    expirations_[key] = int(time)
                except Exception as e:
                    raise AttributeError(f"Expiration value type error, {type(expirations_[key])} was given") from e
        return expirations_

    def get_timeout(self, key):
        try:
            return self.expirations[key]
        except KeyError:
            warnings.warn("Expiration manager received an unknown expiration token, returned 0 by default", PyotPipelineWarning)
            return 0

    def __iter__(self):
        return iter(self.expirations)

    def __len__(self):
        return len(self.expirations)



from datetime import datetime
from typing import TYPE_CHECKING

from pyot.conf.model import models
from pyot.core.functional import empty
from .base import PyotCore

if TYPE_CHECKING:
    from .championmastery import ChampionMasteries
    from .league import SummonerLeague
    from .thirdpartycode import ThirdPartyCode
    from .clash import ClashPlayers
    from .spectator import CurrentGame
    from .profileicon import ProfileIcon
    from ..riot.account import Account
    from .match import MatchHistory


# PYOT CORE OBJECTS

class Summoner(PyotCore):
    name: str
    id: str
    account_id: str
    level: int
    puuid: str
    profile_icon_id: int
    revision_date_millis: int

    class Meta(PyotCore.Meta):
        rules = {
            "summoner_v4_by_puuid": ["puuid"],
            "summoner_v4_by_id": ["id"],
            "summoner_v4_by_account_id": ["account_id"],
            "summoner_v4_by_name": ["name"],
        }
        renamed = {"summoner_level": "level", "revision_date": "revision_date_millis"}

    def __init__(self, id: str = empty, account_id: str = empty, name: str = empty, puuid: str = empty, platform: str = models.lol.DEFAULT_PLATFORM):
        self.initialize(locals())

    @property
    def revision_date(self) -> datetime:
        return datetime.fromtimestamp(self.revision_date_millis//1000)

    @property
    def champion_masteries(self) -> "ChampionMasteries":
        from .championmastery import ChampionMasteries
        return ChampionMasteries(summoner_id=self.id, platform=self.platform)

    @property
    def league_entries(self) -> "SummonerLeague":
        from .league import SummonerLeague
        return SummonerLeague(summoner_id=self.id, platform=self.platform)

    @property
    def third_party_code(self) -> "ThirdPartyCode":
        from .thirdpartycode import ThirdPartyCode
        return ThirdPartyCode(summoner_id=self.id, platform=self.platform)

    @property
    def clash_players(self) -> "ClashPlayers":
        from .clash import ClashPlayers
        return ClashPlayers(summoner_id=self.id, platform=self.platform)

    @property
    def current_game(self) -> "CurrentGame":
        from .spectator import CurrentGame
        return CurrentGame(summoner_id=self.id, platform=self.platform)

    @property
    def profile_icon(self) -> "ProfileIcon":
        from .profileicon import ProfileIcon
        return ProfileIcon(id=self.profile_icon_id)

    @property
    def account(self) -> "Account":
        from ..riot.account import Account
        return Account(puuid=self.puuid).using(self.metapipeline.name)

    @property
    def match_history(self) -> "MatchHistory":
        from .match import MatchHistory
        return MatchHistory(puuid=self.puuid, region=self.region)



from typing import Any
import atexit
import pickle

from diskcache import FanoutCache
from asgiref.sync import sync_to_async
from pyot.core.exceptions import NotFound
from pyot.pipeline.token import PipelineToken
from pyot.pipeline.expiration import ExpirationManager
from pyot.utils.logging import LazyLogger

from .base import Store, StoreType


LOGGER = LazyLogger(__name__)


class DiskCache(Store):

    type = StoreType.CACHE

    def __init__(self, game: str, directory: str, expirations: Any = None, log_level: int = 0, **kwargs) -> None:
        self.game = game
        if "timeout" not in kwargs:
            kwargs["timeout"] = 1
        self.cache = FanoutCache(directory=directory, **kwargs)
        self.alias = str(directory).split("/")[-1] if "/" in str(directory) else str(directory).split("\\")[-1]
        self.expirations = ExpirationManager(game, expirations)
        self.log_level = log_level
        atexit.register(self.cache.close)

    async def set(self, token: PipelineToken, value: Any, **kwargs) -> None:
        timeout = self.expirations.get_timeout(token.method)
        if timeout != 0:
            if timeout == -1:
                timeout = None
            await sync_to_async(self.cache.set)(token.value, pickle.dumps(value), timeout)
            LOGGER.log(self.log_level, f"[pyot.stores.diskcache:DiskCache#{self.alias}] SET {token.value}")

    async def get(self, token: PipelineToken, **kwargs) -> Any:
        timeout = self.expirations.get_timeout(token.method)
        if timeout == 0:
            raise NotFound(token.value)
        item = await sync_to_async(self.cache.get)(token.value)
        if item is None:
            raise NotFound(token.value)
        LOGGER.log(self.log_level, f"[pyot.stores.diskcache:DiskCache#{self.alias}] GET {token.value}")
        return pickle.loads(item)

    async def delete(self, token: PipelineToken, **kwargs) -> None:
        await sync_to_async(self.cache.delete)(token.value)
        LOGGER.log(self.log_level, f"[pyot.stores.diskcache:DiskCache#{self.alias}] DELETE {token.value}")

    async def contains(self, token: PipelineToken, **kwargs) -> bool:
        item = await sync_to_async(self.cache.get)(token.value)
        if item is None:
            return False
        return True

    async def clear(self, **kwargs):
        await sync_to_async(self.cache.clear)()
        LOGGER.log(self.log_level, f"[pyot.stores.diskcache:DiskCache#{self.alias}] Store cleared successfully")




from .engines import ModelsDocEngine, UtilsDocEngine, ExamplesDocEngine


def build_docs():
    ModelsDocEngine().run()
    UtilsDocEngine().run()
    ExamplesDocEngine().run()


if __name__ == '__main__':
    build_docs()



from typing import Any, List


class PyotDocTypeSerializer:

    def __init__(self, type: Any, relative_paths: List[str]) -> None:
        if type is None:
            self.data = None
            return
        self.data = str(type)
        self.data = self.data.replace("typing.", "")
        self.data = self.data.replace("<class ", "")
        self.data = self.data.replace(">", "")
        self.data = self.data.replace("'", "")
        for relative_path in relative_paths:
            if self.data.startswith(relative_path):
                self.data = self.data[len(relative_path):]



from abc import ABC
from typing import Dict, List, Mapping
import warnings

from pyot.core.warnings import PyotConfWarning
from pyot.pipeline.core import Pipeline
from pyot.utils.importlib import import_variable

from .utils import ConfDict


AVAILABLE_MODELS = {"lol", "tft", "lor", "val"}

pipelines: ConfDict[str, Pipeline] = ConfDict(Pipeline, False)


def activate_pipeline(model: str):

    keys = {"name", "stores", "default"}

    if model not in AVAILABLE_MODELS:
        raise ValueError(f"Invalid model, '{model}' was given")

    def build_pipeline(pipeline_configs: List[Dict]):
        stores = []
        for config in pipeline_configs:
            store_cls = import_variable(config.pop("backend"))
            config.update({"game": model.lower()})
            store = store_cls(**config)
            stores.append(store)
        return stores

    def wrapper(cls: "PipelineConf"):
        for key in keys:
            if hasattr(cls, key):
                continue
            raise ValueError(f"Missing value for '{key}' in {cls} conf")
        pipeline = Pipeline(model, cls.name, build_pipeline(cls.stores))
        if cls.name in AVAILABLE_MODELS and (not cls.default or not cls.name == model):
            raise ValueError("Pipeline name must be different than model's name or the same as the subscripted model if set to default")
        if model in pipelines:
            warnings.warn(f"An attempt to activate pipeline '{cls.name}' was ignored due to default pipeline of model already active", PyotConfWarning)
            return cls
        if cls.name in pipelines:
            warnings.warn(f"An attempt to activate pipeline '{cls.name}' was ignored due to pipeline already active", PyotConfWarning)
            return cls
        if cls.default:
            pipelines[model] = pipeline
        pipelines[cls.name] = pipeline
        return cls

    return wrapper


class PipelineConf(ABC):

    name: str
    stores: List[Mapping[str, Dict]]
    default: bool = False



from typing import Any, Dict


class PipelineToken:
    value: str
    hashval: str
    model: str
    server: str
    method: str
    params: Dict[str, str]
    queries: Dict[str, Any]

    def __init__(self, model: str, server: str, method: str, params: Dict[str, str], queries: Dict[str, Any]):
        self.model = model
        self.server = server
        self.method = method
        self.params = params
        self.queries = queries
        if self.server:
            self.parent = f"{self.model}/{self.server}/{self.method}"
        else:
            self.parent = f"{self.model}/{self.method}"
        self.value = f"{self.parent}{self.parse_params(self.params)}{self.parse_queries(self.queries)}"
        self.hashval = hash((self.model, self.server, self.method, str(self.params), str(self.queries)))

    def __hash__(self):
        return self.hashval

    @staticmethod
    def parse_params(dic: Dict) -> str:
        res = "/".join(str(v) for v in dic.values())
        return "/" + res if res else ""

    @staticmethod
    def parse_queries(dic: Dict) -> str:
        res = "&".join(f"{k}={v}" for k, v in dic.items())
        return "?" + res if res else ""

    def dict(self) -> dict:
        return {
            "model": self.model,
            "server": self.server,
            "method": self.method,
            "params": self.params,
            "queries": self.queries,
        }

    @classmethod
    def load(cls, dic) -> 'PipelineToken':
        return cls(dic["model"], dic["server"], dic["method"], dic["params"], dic["queries"])



import asyncio
import contextlib
from datetime import datetime, timedelta

from ..aiohttp import SafeClientSession
from ..functools import async_property


class ChampionKeysCache:

    def __init__(self) -> None:
        self._is_locked = False
        self.cached_data = {
            "id_by_key": {},
            "id_by_name": {},
            "key_by_id": {},
            "key_by_name": {},
            "name_by_id": {},
            "name_by_key": {},
        }
        self.last_updated = datetime.now() - timedelta(days=1)

    def __str__(self) -> str:
        return 'ChampionKeysCache()'

    @property
    @contextlib.asynccontextmanager
    async def _lock(self):
        try:
            while self._is_locked:
                await asyncio.sleep(0.01)
            self._is_locked = True
            yield None
        finally:
            self._is_locked = False

    @async_property
    async def data(self):
        if datetime.now() - self.last_updated < timedelta(hours=3):
            return self.cached_data
        async with self._lock:
            if datetime.now() - self.last_updated < timedelta(hours=3):
                return self.cached_data
            url = "https://raw.communitydragon.org/pbe/plugins/rcp-be-lol-game-data/global/en_gb/v1/champion-summary.json"
            async with SafeClientSession() as session:
                response = await session.get(url)
                if not (response and response.status == 200):
                    raise RuntimeError(f"Failed to pull champion summary ({response.status})")
                dic = await response.json(encoding="utf-8")
            data = {
                "id_by_key": {},
                "id_by_name": {},
                "key_by_id": {},
                "key_by_name": {},
                "name_by_id": {},
                "name_by_key": {},
            }
            for champ in dic:
                if champ["id"] == -1:
                    continue
                data["id_by_key"][champ["alias"]] = champ["id"]
                data["id_by_name"][champ["name"]] = champ["id"]
                data["key_by_id"][champ["id"]] = champ["alias"]
                data["key_by_name"][champ["name"]] = champ["alias"]
                data["name_by_id"][champ["id"]] = champ["name"]
                data["name_by_key"][champ["alias"]] = champ["name"]
            self.cached_data = data
            self.last_updated = datetime.now()
        return self.cached_data


champion_keys_cache = ChampionKeysCache()


async def id_by_key(value: str) -> int:
    '''Get champion id by key'''
    data = await champion_keys_cache.data
    return data["id_by_key"][value]


async def id_by_name(value: str) -> int:
    '''Get champion id by name'''
    data = await champion_keys_cache.data
    return data["id_by_name"][value]


async def key_by_id(value: int) -> str:
    '''Get champion key by id'''
    data = await champion_keys_cache.data
    return data["key_by_id"][value]


async def key_by_name(value: str) -> str:
    '''Get champion key by name'''
    data = await champion_keys_cache.data
    return data["key_by_name"][value]


async def name_by_id(value: int) -> str:
    '''Get champion name by id'''
    data = await champion_keys_cache.data
    return data["name_by_id"][value]


async def name_by_key(value: str) -> str:
    '''Get champion name by key'''
    data = await champion_keys_cache.data
    return data["name_by_key"][value]



import warnings

from pyot.conf.utils import import_confs
from pyot.core.warnings import PyotConfWarning


try:
    from django.core.exceptions import ImproperlyConfigured
    try:
        from django.conf import settings

        conf_paths = settings.PYOT_CONFS
        DJANGO_ENABLED = True
    except ImproperlyConfigured:
        DJANGO_ENABLED = False
    except AttributeError:
        warnings.warn("Django install detected, but could not find `PYOT_CONFS` settings variable", PyotConfWarning)
        DJANGO_ENABLED = False
except (ImportError, ModuleNotFoundError, NameError):
    DJANGO_ENABLED = False


def activate():
    if DJANGO_ENABLED:
        import_confs(conf_paths)



from typing import Dict, List, Iterator, TYPE_CHECKING

from pyot.conf.model import models
from pyot.core.functional import empty
from .base import PyotCore, PyotStatic

if TYPE_CHECKING:
    from ..riot.account import Account


# PYOT STATIC OBJECTS

class LeaderboardPlayerData(PyotStatic):
    puuid: str
    game_name: str
    tag_line: str
    leaderboard_rank: int
    ranked_rating: int
    number_of_wins: int
    competitive_tier: int

    @property
    def account(self) -> "Account":
        from ..riot.account import Account
        return Account(puuid=self.puuid).using(self.metapipeline.name)


class LeaderboardTierDetailData(PyotStatic):
    ranked_rating_threshold: int
    starting_page: int
    starting_index: int


# PYOT CORE OBJECTS

class Leaderboard(PyotCore):
    act_id: str
    total_players: int
    players: List[LeaderboardPlayerData]
    immortal_starting_page: int
    immortal_starting_index: int
    top_tier_rr_threshold: int
    tier_details: Dict[str, LeaderboardTierDetailData]
    start_index: int
    query_str: str
    shard: str

    class Meta(PyotCore.Meta):
        rules = {"ranked_v1_leaderboards": ["act_id"]}
        renamed = {"query": "query_str"}

    def __init__(self, act_id: str = empty, platform: str = models.val.DEFAULT_PLATFORM):
        self.initialize(locals())

    def __getitem__(self, item):
        if not isinstance(item, int):
            return super().__getitem__(item)
        return self.players[item]

    def __iter__(self) -> Iterator[LeaderboardPlayerData]:
        return iter(self.players)

    def __len__(self) -> int:
        return len(self.players)

    def query(self, size: int = 200, start_index: int = 0):
        '''Set query request parameters.'''
        super()._place_query(locals())
        return self



from pyot.core.objects import PyotCoreBase, PyotStaticBase
from pyot.conf.pipeline import pipelines


MODULE_REPR = 'Legends of Runeterra'


class PyotRouting:

    _regions = {"americas", "europe", "asia", "esports", "sea", "apac"}


class PyotCore(PyotRouting, PyotCoreBase):

    class Meta(PyotCoreBase.Meta):
        pipeline = pipelines.lor


class PyotStatic(PyotRouting, PyotStaticBase):

    class Meta(PyotStaticBase.Meta):
        pipeline = pipelines.lor


bases = (PyotCore, PyotStatic)



from lor_deckcodes.models import CardCodeAndCount


def batch_to_ccac(batch):
    '''
    Converts a Batch object to CardCodeAndCount object.
    '''
    return CardCodeAndCount.from_card_string(str(batch))

def ccac_to_batch(ccac: CardCodeAndCount):
    '''
    Converts a CardCodeAndCount object to Batch object.
    '''
    from pyot.models.lor import Batch
    return Batch(code=ccac.card_code, count=ccac.count)



from datetime import datetime, timedelta
from typing import List, Iterator, Dict, TYPE_CHECKING, Tuple, Union

from pyot.conf.model import models
from pyot.core.functional import lazy_property, empty

from .base import PyotCore, PyotStatic

if TYPE_CHECKING:
    from .item import Item
    from .champion import Champion
    from .merakichampion import MerakiChampion
    from .merakiitem import MerakiItem
    from .rune import Rune
    from .spell import Spell
    from .summoner import Summoner


# PYOT STATIC OBJECTS

class MatchBanData(PyotStatic):
    champion_id: int
    pick_turn: int

    @property
    def champion(self) -> "Champion":
        from .champion import Champion
        return Champion(id=self.champion_id)

    @property
    def meraki_champion(self) -> "MerakiChampion":
        from .merakichampion import MerakiChampion
        return MerakiChampion(id=self.champion_id)


class MatchObjectiveDetailData(PyotStatic):
    first: bool
    kills: int


class MatchObjectiveData(PyotStatic):
    baron: MatchObjectiveDetailData
    champion: MatchObjectiveDetailData
    dragon: MatchObjectiveDetailData
    inhibitor: MatchObjectiveDetailData
    rift_herald: MatchObjectiveDetailData
    tower: MatchObjectiveDetailData


class MatchTeamData(PyotStatic):
    id: int
    win: bool
    bans: List[MatchBanData]
    objectives: MatchObjectiveData

    class Meta(PyotStatic.Meta):
        renamed = {"team_id": "id"}

    @property
    def participants(self) -> List["MatchParticipantData"]:
        self.metaroot: Match
        return [p for p in self.metaroot.info.participants if p.team_id == self.id]


class MatchPerkSelectionData(PyotStatic):
    perk: int
    var1: int
    var2: int
    var3: int


class MatchPerkStyleData(PyotStatic):
    description: str
    selections: List[MatchPerkSelectionData]
    style: int


class MatchStatPerkData(PyotStatic):
    offense: int
    flex: int
    defense: int


class MatchPerkData(PyotStatic):
    stat_perks: MatchStatPerkData
    styles: List[MatchPerkStyleData]


class TimelineChampionStatData(PyotStatic):
    ability_haste: int
    ability_power: int
    armor: int
    armor_pen: int
    armor_pen_percent: int
    attack_damage: int
    attack_speed: int
    bonus_armor_pen_percent: int
    bonus_magic_pen_percent: int
    cc_reduction: int
    cooldown_reduction: int
    health: int
    health_max: int
    health_regen: int
    lifesteal: int
    magic_pen: int
    magic_pen_percent: int
    magic_resist: int
    movement_speed: int
    omnivamp: int
    physical_vamp: int
    power: int
    power_max: int
    power_regen: int
    spell_vamp: int


class TimelinePositionData(PyotStatic):
    x: int
    y: int


class TimelineDamageStatData(PyotStatic):
    magic_damage_done: int
    magic_damage_done_to_champions: int
    magic_damage_taken: int
    physical_damage_done: int
    physical_damage_done_to_champions: int
    physical_damage_taken: int
    total_damage_done: int
    total_damage_done_to_champions: int
    total_damage_taken: int
    true_damage_done: int
    true_damage_done_to_champions: int
    true_damage_taken: int


class TimelineParticipantFrameData(PyotStatic):
    champion_stats: TimelineChampionStatData
    current_gold: int
    damage_stats: TimelineDamageStatData
    gold_per_second: int
    jungle_minions_killed: int
    level: int
    minions_killed: int
    participant_id: int
    position: TimelinePositionData
    time_enemy_spent_controlled_millis: int
    total_gold: int
    xp: int

    class Meta(PyotStatic.Meta):
        renamed = {"time_enemy_spent_controlled": "time_enemy_spent_controlled_millis"}

    @property
    def time_enemy_spent_controlled(self) -> timedelta:
        return timedelta(milliseconds=self.time_enemy_spent_controlled_millis)


class TimelineVictimDamageData(PyotStatic):
    basic: bool
    magic_damage: int
    name: str
    participant_id: int
    physical_damage: int
    spell_name: str
    spell_slot: int
    true_damage: int
    type: str


class TimelineEventData(PyotStatic):
    actual_start_time_millis: int
    ascended_type: str
    assisting_participant_ids: List[int]
    after_id: int
    before_id: int
    bounty: int
    building_type: str
    creator_id: int
    event_type: str
    game_id: int
    gold_gain: int
    item_id: int
    kill_streak_length: int
    kill_type: str
    killer_id: int
    killer_team_id: int
    lane_type: str
    level: int
    level_up_type: str
    monster_type: str
    monster_sub_type: str
    multi_kill_length: int
    name: str
    participant_id: int
    point_captured: str
    position: TimelinePositionData
    real_timestamp: int
    skill_slot: int
    shutdown_bounty: int
    team_id: int
    timestamp: int
    transform_type: str
    type: str
    tower_type: str
    victim_id: int
    victim_damage_dealt: List[TimelineVictimDamageData]
    victim_damage_received: List[TimelineVictimDamageData]
    ward_type: str
    winning_team: int

    class Meta(PyotStatic.Meta):
        raws = {"assisting_participant_ids"}
        renamed = {"actual_start_time": "actual_start_time_millis"}

    @property
    def actual_start_time(self) -> timedelta:
        return timedelta(milliseconds=self.actual_start_time_millis)

    @property
    def time(self) -> timedelta:
        return timedelta(milliseconds=self.timestamp)

    @property
    def real_time(self) -> datetime:
        return datetime.fromtimestamp(self.timestamp // 1000)

    @property
    def after_item(self) -> "Item":
        from .item import Item
        return Item(id=self.after_id)

    @property
    def item(self) -> "Item":
        from .item import Item
        return Item(id=self.item_id)

    @property
    def before_item(self) -> "Item":
        from .item import Item
        return Item(id=self.before_id)

    @property
    def meraki_after_item(self) -> "MerakiItem":
        from .merakiitem import MerakiItem
        return MerakiItem(id=self.after_id)

    @property
    def meraki_item(self) -> "MerakiItem":
        from .merakiitem import MerakiItem
        return MerakiItem(id=self.item_id)

    @property
    def meraki_before_item(self) -> "MerakiItem":
        from .merakiitem import MerakiItem
        return MerakiItem(id=self.before_id)


class MatchParticipantData(PyotStatic):
    id: int
    assists: int
    baron_kills: int
    basic_pings: int
    bounty_level: int
    champ_experience: int
    champ_level: int
    champion_id: int
    champion_name: str
    champion_transform: int
    consumables_purchased: int
    damage_dealt_to_buildings: int
    damage_dealt_to_objectives: int
    damage_dealt_to_turrets: int
    damage_self_mitigated: int
    deaths: int
    detector_wards_placed: int
    double_kills: int
    dragon_kills: int
    eligible_for_progression: bool
    first_blood_assist: bool
    first_blood_kill: bool
    first_tower_assist: bool
    first_tower_kill: bool
    game_ended_in_early_surrender: bool
    game_ended_in_surrender: bool
    gold_earned: int
    gold_spent: int
    individual_position: str
    inhibitor_kills: int
    inhibitor_takedowns: int
    inhibitors_lost: int
    item0: int
    item1: int
    item2: int
    item3: int
    item4: int
    item5: int
    item6: int
    items_purchased: int
    killing_sprees: int
    kills: int
    lane: str
    largest_critical_strike: int
    largest_killing_spree: int
    largest_multi_kill: int
    longest_time_spent_living_secs: int
    magic_damage_dealt: int
    magic_damage_dealt_to_champions: int
    magic_damage_taken: int
    neutral_minions_killed: int
    nexus_kills: int
    nexus_takedowns: int
    nexus_lost: int
    objectives_stolen: int
    objectives_stolen_assists: int
    penta_kills: int
    perks: MatchPerkData
    physical_damage_dealt: int
    physical_damage_dealt_to_champions: int
    physical_damage_taken: int
    profile_icon_id: int
    puuid: str
    quadra_kills: int
    riot_id_name: str
    riot_id_tagline: str
    role: str
    sight_wards_bought_in_game: int
    spell1_casts: int
    spell2_casts: int
    spell3_casts: int
    spell4_casts: int
    summoner1_casts: int
    summoner1_id: int
    summoner2_casts: int
    summoner2_id: int
    summoner_id: str
    summoner_level: int
    summoner_name: str
    team_early_surrendered: bool
    team_id: int
    team_position: str
    time_ccing_others_secs: int
    time_played_secs: int
    total_damage_dealt: int
    total_damage_dealt_to_champions: int
    total_damage_shielded_on_teammates: int
    total_damage_taken: int
    total_heal: int
    total_heals_on_teammates: int
    total_minions_killed: int
    total_time_cc_dealt_secs: int
    total_time_spent_dead_secs: int
    total_units_healed: int
    triple_kills: int
    true_damage_dealt: int
    true_damage_dealt_to_champions: int
    true_damage_taken: int
    turret_kills: int
    turret_takedowns: int
    turrets_lost: int
    unreal_kills: int
    vision_score: int
    vision_wards_bought_in_game: int
    wards_killed: int
    wards_placed: int
    challenges: Dict[str, float]
    frames: List[TimelineParticipantFrameData]
    events: List[TimelineEventData]
    win: bool

    class Meta(PyotStatic.Meta):
        raws = {'challenges'}
        renamed = {
            "participant_id": "id", "profile_icon": "profile_icon_id", "time_c_cing_others": "time_ccing_others_secs",
            "total_time_cc_dealt": "total_time_cc_dealt_secs", "total_time_spent_dead": "total_time_spent_dead_secs",
            "time_played": "time_played_secs", "longest_time_spent_living": "longest_time_spent_living_secs",
        }

    @property
    def total_time_cc_dealt(self) -> timedelta:
        return timedelta(seconds=self.total_time_cc_dealt_secs)

    @property
    def total_time_spent_dead(self) -> timedelta:
        return timedelta(seconds=self.total_time_spent_dead_secs)

    @property
    def time_ccing_others(self) -> timedelta:
        return timedelta(seconds=self.time_ccing_others_secs)

    @property
    def time_played(self) -> timedelta:
        return timedelta(seconds=self.time_played_secs)

    @property
    def longest_time_spent_living(self) -> timedelta:
        return timedelta(seconds=self.longest_time_spent_living_secs)

    @lazy_property
    def rune_ids(self) -> List[int]:
        ids = []
        for style in self.perks.styles:
            for selection in style.selections:
                ids.append(selection.perk)
        return ids

    @lazy_property
    def item_ids(self) -> List[int]:
        return [self.item0, self.item1, self.item2, self.item3, self.item4, self.item5, self.item6]

    @property
    def items(self) -> List["Item"]:
        from .item import Item
        return [Item(id=id_) for id_ in self.item_ids]

    @property
    def meraki_items(self) -> List["MerakiItem"]:
        from .merakiitem import MerakiItem
        return [MerakiItem(id=id_) for id_ in self.item_ids]

    @lazy_property
    def stat_rune_ids(self) -> List[int]:
        o = self.perks.stat_perks
        return [o.offense, o.flex, o.defense]

    @lazy_property
    def rune_primary_style(self) -> int:
        return next(style.style for style in self.perks.styles if style.description == "primaryStyle")

    @lazy_property
    def rune_sub_style(self) -> int:
        return next(style.style for style in self.perks.styles if style.description == "subStyle")

    @property
    def runes(self) -> List["Rune"]:
        from .rune import Rune
        return [Rune(id=id_) for id_ in self.rune_ids]

    @lazy_property
    def spell_ids(self) -> List[int]:
        return [self.summoner1_id, self.summoner2_id]

    @property
    def spells(self) -> List["Spell"]:
        from .spell import Spell
        return [Spell(id=id_) for id_ in self.spell_ids]

    @property
    def summoner(self) -> "Summoner":
        from .summoner import Summoner
        return Summoner(puuid=self.puuid, platform=self.metaroot.platform)


class MatchMetaData(PyotStatic):
    match_id: str
    data_version: str
    participant_puuids: List[str]

    class Meta(PyotStatic.Meta):
        raws = {"participant_puuids"}
        renamed = {"participants": "participant_puuids"}

    @property
    def participants(self) -> "Summoner":
        from .summoner import Summoner
        return [Summoner(puuid=puuid, platform=self.metaroot.platform) for puuid in self.participant_puuids]


class MatchInfoData(PyotStatic):
    game_id: int
    creation_millis: int
    duration_units: int
    start_millis: int
    end_millis: int
    mode: str
    name: str
    type: str
    version: str
    platform: str
    map_id: int
    queue_id: int
    tournament_code: str
    participants: List[MatchParticipantData]
    teams: List[MatchTeamData]

    class Meta(PyotStatic.Meta):
        renamed = {
            "game_creation": "creation_millis", "game_duration": "duration_units", "game_mode": "mode", "game_name": "name",
            "game_start_timestamp": "start_millis", "game_type": "type", "game_version": "version", "platform_id": "platform",
            "game_end_timestamp": "end_millis",
        }

    @property
    def creation(self) -> datetime:
        return datetime.fromtimestamp(self.creation_millis // 1000)

    @property
    def duration(self) -> timedelta:
        return timedelta(seconds=self.duration_secs)

    @property
    def start(self) -> datetime:
        return datetime.fromtimestamp(self.start_millis // 1000)

    @property
    def end(self) -> datetime:
        return datetime.fromtimestamp(self.end_millis // 1000)

    @property
    def duration_millis(self) -> int:
        if self.duration_units < 25200:
            return self.duration_units * 1000
        return self.duration_units

    @property
    def duration_secs(self) -> int:
        if self.duration_units > 25200:
            return self.duration_units / 1000
        return self.duration_units

    @property
    def start(self) -> datetime:
        return datetime.fromtimestamp(self.start_millis // 1000)


class TimelineFrameData(PyotStatic):
    events: List[TimelineEventData]
    participant_frames: List[TimelineParticipantFrameData]
    timestamp: int

    @property
    def time(self) -> timedelta:
        return timedelta(milliseconds=self.timestamp)


class TimelineParticipantData(PyotStatic):
    id: int
    puuid: str

    class Meta(PyotStatic.Meta):
        renamed = {"participant_id": "id"}

    @property
    def summoner(self) -> "Summoner":
        from .summoner import Summoner
        return Summoner(puuid=self.puuid, platform=self.metaroot.platform)


class TimelineInfoData(PyotStatic):
    frame_interval_millis: int
    frames: List[TimelineFrameData]
    game_id: int
    participants: List[TimelineParticipantData]

    class Meta(PyotStatic.Meta):
        renamed = {"frame_interval": "frame_interval_millis"}

    @property
    def frame_interval(self) -> timedelta:
        return timedelta(milliseconds=self.frame_interval_millis)


# PYOT CORE OBJECTS

class Match(PyotCore):
    metadata: MatchMetaData
    info: MatchInfoData
    id: str

    class Meta(PyotCore.Meta):
        rules = {"match_v5_match": ["id"]}

    def __init__(self, id: str = empty, region: str = models.lol.DEFAULT_REGION):
        self.initialize(locals())

    @property
    def platform(self) -> str:
        try:
            return self.info.platform
        except AttributeError:
            return super().platform

    def feed_timeline(self, timeline: "Timeline", include_assisted=False, include_victim=False):
        '''Parse the given `Timeline` object's frames and events into this match's participants.
        - `include_assisted`: Include frames and events where the participants are scoring an assist.
        - `include_victim`: Include frames and events where the participants are victims.
        '''
        participants: Dict[int, MatchParticipantData] = {}
        for participant in self.info["participants"]:
            participants[participant["participantId"]] = participant
            participant["frames"] = []
            participant["events"] = []
        idkeys = ["participantId", "creatorId", "killerId"]
        for frame in timeline.info.frames:
            for participant_frame in frame["participantFrames"]:
                participants[participant_frame["participantId"]]["frames"].append(participant_frame)
            for event in frame["events"]:
                pid = None
                for idkey in idkeys:
                    pid = event.get(idkey, None)
                    if pid is not None:
                        break
                if not pid:
                    continue
                participants[pid]["events"].append(event)
                if include_assisted and "assistingParticipantIds" in event:
                    for apid in event["assistingParticipantIds"]:
                        participants[apid]["events"].append(event)
                if include_victim and "victimId" in event:
                    participants[event["victimId"]]["events"].append(event)


class Timeline(PyotCore):
    metadata: MatchMetaData
    info: TimelineInfoData
    id: str

    class Meta(PyotCore.Meta):
        rules = {"match_v5_timeline": ["id"]}

    def __init__(self, id: str = empty, region: str = models.lol.DEFAULT_REGION):
        self.initialize(locals())

    def transform(self, data):
        data = data.copy()
        data["info"] = data["info"].copy()
        data["info"]["frames"] = data["info"]["frames"].copy()
        for i, f in enumerate(data["info"]["frames"]):
            data["info"]["frames"][i] = f.copy()
            if f["participantFrames"]:
                data["info"]["frames"][i]["participantFrames"] = list(sorted(f["participantFrames"].values(), key=lambda p: p["participantId"]))
        return data

    @property
    def platform(self) -> "str":
        try:
            return self.metadata.match_id.split("_")[0]
        except AttributeError:
            return super().platform


class MatchHistory(PyotCore):
    ids: List[str]
    puuid: str

    class Meta(PyotCore.Meta):
        rules = {"match_v5_matches": ["puuid"]}
        raws = {"ids"}

    def __init__(self, puuid: str = empty, region: str = models.lol.DEFAULT_REGION):
        self.initialize(locals())

    def __getitem__(self, item) -> Match:
        if not isinstance(item, int):
            return super().__getitem__(item)
        return Match(id=self.ids[item], region=self.region)

    def __iter__(self) -> Iterator[Match]:
        return iter(self.matches)

    def __len__(self) -> int:
        return len(self.ids)

    @property
    def matches(self) -> List[Match]:
        return [Match(id=id_, region=self.region) for id_ in self.ids]

    @property
    def timelines(self) -> List[Timeline]:
        return [Timeline(id=id_, region=self.region) for id_ in self.ids]

    @property
    def match_timelines(self) -> List[Tuple[Match, Timeline]]:
        return [(Match(id=id_, region=self.region), Timeline(id=id_, region=self.region)) for id_ in self.ids]

    def query(self, start: int = 0, count: int = 20, queue: int = empty, type: str = empty, start_time: Union[int, datetime] = empty, end_time: Union[int, datetime] = empty):
        '''Set query request parameters.'''
        if isinstance(start_time, datetime):
            start_time = int(start_time.timestamp())
        if isinstance(end_time, datetime):
            end_time = int(end_time.timestamp())
        super()._place_query(locals())
        return self

    def transform(self, data) -> Dict:
        return {"ids": data}



import warnings


class PyotConfWarning(UserWarning):
    pass


class PyotPipelineWarning(RuntimeWarning):
    pass


class PyotStoreWarning(RuntimeWarning):
    pass


class PyotRuntimeWarning(RuntimeWarning):
    pass


class PyotResourceWarning(ResourceWarning):
    pass


warnings.simplefilter("default", PyotResourceWarning)
